from pyessv._ws.handlers.heartbeat import HeartbeatRequestHandler
from pyessv._ws.handlers.retrieve import RetrieveRequestHandler
from pyessv._ws.handlers.validate_identifier import ValidateIdentifierHandler
from pyessv._ws.handlers.validate_identifier_set import ValidateIdentifierSetHandler



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._parsers.cmip6_dataset_id.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Encapsulates parsing of a CMIP6 dataset identifier.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>

"""
from pyessv._factory import create_template_parser
from pyessv._constants import PARSING_STRICTNESS_1



# Template extracted from esgf ini file (for reference purpose only).
_INI_PATTERN = '%(activity_id)s.%(mip_era)s.%(target_mip)s.%(institution_id)s.%(source_id)s.%(realm)s.%(frequency)s.%(variable_id)s.%(grid_label)s'

# Template that identifiers must conform to.
_TEMPLATE = 'input4MIPs.CMIP6.{}.{}.{}.{}.{}.{}.{}'

# Collections injected into template.
_COLLECTIONS = (
    'wcrp:input4mips:target-mip',
    'wcrp:input4mips:institution-id',
    'wcrp:input4mips:source-id',
    'wcrp:input4mips:realm',
    'wcrp:input4mips:frequency',
    'wcrp:input4mips:variable-id',
    'wcrp:input4mips:grid-label',
    )

# Instantiated & cached parser instance.
_PARSER = None


def parse(identifier):
    """Parses a CMIP6 dataset identifier.

    """
    print identifier
    # Instantiate parser JIT.
    global _PARSER
    if _PARSER is None:
        _PARSER = create_template_parser(_TEMPLATE, _COLLECTIONS, PARSING_STRICTNESS_1)

    # Strip version suffix.
    if '#' in identifier:
      identifier = identifier.split('#')[0]

    return _PARSER.parse(identifier)



# -*- coding: utf-8 -*-
"""

.. module:: utils.constants.py
   :license: GPL/CeCIL
   :platform: Unix
   :synopsis: ES-DOC Errata - web-service constants.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
# Default endpoint.
DEFAULT_ENDPOINT = r'/1/retrieve'

# HTTP CORS header.
HTTP_HEADER_Access_Control_Allow_Origin = "Access-Control-Allow-Origin"

# Processing error HTTP response codes.
HTTP_RESPONSE_BAD_REQUEST_ERROR = 400
HTTP_RESPONSE_SERVER_ERROR = 500



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._accessors.wcrp_cmip6.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Encapsulates access to WCRP CMIP6 vocabs.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import pyessv



def _get_scope():
    """Returns target scope.

    """
    return pyessv.WCRP.cmip6


def get_institutes():
    """Returns set of participating institutes.

    """
    scope = _get_scope()

    return scope.institution_id


def get_institute_sources(institution_id):
    """Returns collection of source identifiers related to an institute.

    :param str institution_id: ID of an institute.

    :returns: Collection of source identifiers related to an institute.
    :rtype: list

    """
    scope = _get_scope()

    try:
        institution_id = institution_id.canonical_name
    except AttributeError:
        pass

    def _is_related(source_id):
        return institution_id in [i.lower() for i in source_id.data['institution_id']]

    return [i for i in scope.source_id if _is_related(i)]


def get_source_realms(source_id=None):
    """Returns collection of realms related to a source identifier.

    :param str source_id: ID of a source.

    :returns: Collection of realms related to a source-id.
    :rtype: list

    """
    def _is_realized(realm):
        return source_id.data['model_component'][realm.raw_name]['description'] != 'none'

    scope = _get_scope()

    return scope.realm if source_id is None else [i for i in scope.realm if _is_realized(i)]



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._parser.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Encapsulates parsing of names to match vocabulary entities.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>

"""
from pyessv._loader import load
from pyessv._constants import PARSING_NODE_FIELDS
from pyessv._constants import PARSING_STRICTNESS_SET
from pyessv._constants import PARSING_STRICTNESS_0
from pyessv._constants import PARSING_STRICTNESS_1
from pyessv._constants import PARSING_STRICTNESS_2
from pyessv._constants import PARSING_STRICTNESS_3
from pyessv._constants import PARSING_STRICTNESS_4
from pyessv._exceptions import ParsingError
from pyessv._utils.compat import str



def parse(
    namespace,
    strictness=PARSING_STRICTNESS_2,
    field='canonical_name'
    ):
    """Parses a namespace within a vocabulary hierachy.

    :param str namespace: Vocabulary namespace, e.g. wcrp.
    :param int strictness: Strictness level to apply when applying lookup rules.
    :param str field: Term field to return.

    """
    assert strictness in PARSING_STRICTNESS_SET, 'Invalid parsing strictness'
    assert field in PARSING_NODE_FIELDS, 'Invalid field'

    # Set namespace
    ns = str(namespace).strip().split(':')
    assert len(ns) >= 1 and len(ns) <= 4, 'Invalid namespace'
    ns = ns + [None for i in range(4 - len(ns))]

    # Destructure.
    authority, scope, collection, term = ns

    # Set parsing targets.
    targets = [
        _NodeInfo('authority', authority, strictness),
        _NodeInfo('scope', scope, strictness),
        _NodeInfo('collection', collection, strictness),
        _NodeInfo('term', term, strictness)
    ]
    targets = [i for i in targets if i.name is not None]

    # Load parsed nodes.
    for target in targets:
        namespace = [i.get_name(target) for i in targets]
        namespace = [i for i in namespace if i is not None]
        namespace = ":".join(namespace)
        node = load(namespace)
        target.set_node(node)

    return getattr(target.node, field)


class _NodeInfo(object):
    """Information about a node whose name is being parsed.

    """
    def __init__(self, typekey, name, strictness):
        """Instance constructor.

        """
        self.node = None
        self.name = name
        self.strictness = strictness
        self.typekey = typekey


    def get_name(self, target):
        """Gets parsing relative name.

        """
        if self.node is not None:
            return self.node.canonical_name
        elif self == target:
            return str(self.name).strip().lower()


    def set_node(self, node):
        """Sets node returned from archive search.

        """
        if node is None:
            raise ParsingError(self.typekey, self.name)

        # Confirm match based upon the level of parsing strictness perform test.
        matched = False
        if self.strictness == PARSING_STRICTNESS_0:
            if self.name == node.canonical_name:
                matched = True

        elif self.strictness == PARSING_STRICTNESS_1:
            if self.name == node.raw_name:
                matched = True

        elif self.strictness == PARSING_STRICTNESS_2:
            if self.name in {node.canonical_name, node.raw_name}:
                matched = True

        elif self.strictness == PARSING_STRICTNESS_3:
            if self.name in {node.canonical_name, node.raw_name}.union(set(node.alternative_names)):
                matched = True

        elif self.strictness == PARSING_STRICTNESS_4:
            name = str(self.name).strip().lower()
            if name in [i.lower() for i in node.all_names]:
                matched = True

        # Raise parsing error if appropriate.
        if matched == False:
            raise ParsingError(self.typekey, self.name)

        self.node = node



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._model.authority.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: A vocabulary authority, e.g. WGCM.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyessv._constants import NODE_TYPEKEY_AUTHORITY
from pyessv._constants import REGEX_CANONICAL_NAME
from pyessv._model.node import IterableNode
from pyessv._model.scope import Scope
from pyessv._utils.validation import assert_iterable
from pyessv._utils.validation import assert_regex
from pyessv._utils.validation import assert_string



class Authority(IterableNode):
    """An authority assuming responsibity for governance of vocabularies.

    """
    def __init__(self):
        """Instance constructor.

        """
        self.scopes = []
        super(Authority, self).__init__(self.scopes, NODE_TYPEKEY_AUTHORITY)


    @property
    def ancestors(self):
        """Gets ancestors within archive hierarchy.

        """
        return []


    def get_validators(self):
        """Returns set of validators.

        """
        def _canonical_name():
            assert_string(self.canonical_name)
            assert_regex(self.canonical_name, REGEX_CANONICAL_NAME)

        def _scopes():
            assert_iterable(self.scopes, Scope)

        return super(Authority, self).get_validators() + (
            _canonical_name,
            _scopes
            )



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._validation.py
   :copyright: Copyright "December 01, 2016', IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Encpasulates domain model class instance validation.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import datetime
import inspect
import re

from pyessv._constants import NODE_TYPEKEY_SET
from pyessv._constants import GOVERNANCE_STATUS_SET
from pyessv._constants import REGEX_CANONICAL_NAME
from pyessv._model import Node
from pyessv._utils.compat import basestring
from pyessv._utils.compat import urlparse



def is_valid(instance):
    """Gets flag indicating validity status of a domain model class.

    :returns: Validity status of a domain model class.
    :rtype: bool

    """
    return len(validate(instance)) == 0


def get_errors(instance):
    """Returns sorted list of domain model class instance validation errors.

    :returns: Sorted list of domain model class instance validation errors.
    :rtype: list

    """
    return sorted(list(validate(instance)))


def validate(instance):
    """Validates an instance of a domain node.

    :param instance: Sub-class of pyessv.Node

    :returns: Set of instance validation errrors.
    :rtype: set

    """
    assert isinstance(instance, Node), 'Invalid vocabulary type'

    errs = set()
    for validator in instance.get_validators():
        try:
            validator()
        except AssertionError as err:
            typekey = instance.__class__.__name__
            field = validator.__name__[1:]
            val = getattr(instance, field)
            errs.add('{}: invalid {}: [{}]'.format(typekey, field, val))

    return errs



# -*- coding: utf-8 -*-

"""
.. module:: utils.validation.py
   :copyright: @2015 IPSL (http://ipsl.fr)
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Variable validation utilities.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>

"""
import arrow
import requests
import pyessv

from pyessv._ws.utils import config
from pyessv._ws.utils import exceptions



def validate(validator):
    """Validation function decorator.

    """
    def decorate(func):
        """The decorator."""
        def wrapper(*args, **kwargs):
            """The wrapper."""
            validator(*args, **kwargs)
            return func(*args, **kwargs)
        return wrapper
    return decorate


def _raise_value_error(val, var, var_type):
    """Raises a generic value error.

    """
    raise ValueError('{0} [{1}] is an invalid {2}'.format(var, val, var_type))


def validate_bool(val, var):
    """Validates a boolean.

    """
    if val is None:
        raise ValueError('{0} is undefined bool'.format(var))

    try:
        bool(val)
    except ValueError:
        _raise_value_error(val, var, bool)


def validate_enum(enum_values, val, var):
    """Validates an enumerable member.

    """
    if val not in enum_values:
        _raise_value_error(val, var, 'enum-member')


def validate_pyessv_enum(collection_id, val, var):
    """Validates an enumerable member.

    """
    namespace = '{}:{}'.format(collection_id, val)
    try:
        pyessv.parse(namespace)
    except pyessv.ParsingError:
        _raise_value_error(val, var, 'enum-member')


def validate_int(val, var):
    """Validates an integer.

    """
    if val is None:
        raise ValueError('{0} is undefined'.format(var))

    try:
        int(val)
    except ValueError:
        _raise_value_error(val, var, int)


def validate_float(val, var):
    """Validates a flaot.

    """
    if val is None:
        raise ValueError('{0} is undefined'.format(var))

    try:
        float(val)
    except ValueError:
        _raise_value_error(val, var, float)


def validate_date(val, var, date_format=None):
    """Validates a date.

    """
    if val is None:
        raise ValueError('{0} is undefined date'.format(var))

    try:
        if date_format is not None:
            arrow.get(val, date_format)
        else:
            arrow.get(val)
    except arrow.parser.ParserError:
        _raise_value_error(val, var, 'date')


def validate_str(val, var):
    """Validates a string.

    """
    if val is None:
        raise ValueError('{0} is undefined string'.format(var))

    try:
        val = str(val)
    except ValueError:
        _raise_value_error(val, var, str)

    if not len(val):
        raise ValueError('{0} is empty string'.format(var))


def validate_unicode(val, var):
    """Validates a unicode.

    """
    if val is None:
        raise ValueError('{0} is undefined unicode'.format(var))

    try:
        val = unicode(val)
    except ValueError:
        _raise_value_error(val, var, unicode)

    if not len(val):
        raise ValueError('{0} is empty unicode'.format(var))


def validate_iterable(val, var):
    """Validates an iterable.

    """
    if val is None:
        raise ValueError('{0} is undefined iterable'.format(var))

    try:
        iter(val)
    except TypeError:
        _raise_value_error(val, var, iter)


def validate_url(url):
    """Validates that a URL will not return a 404.

    :param str url: The url to validate.

    """
    if url in (None, ''):
        return

    if url.startswith('http'):
        if not config.validate_issue_urls:
            return
        response = requests.head(url)
        if response.status_code in [requests.codes.OK, 302]:
            return
        response = requests.get(url)
        if response.ok:
            return

    raise exceptions.InvalidURLError(url)



# -*- coding: utf-8 -*-

"""
.. module:: test_parser_name.py

   :copyright: @2013 Earth System Documentation (https://es-doc.org)
   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Executes pyessv name parsing tests.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import nose

import pyessv as LIB
import tests.utils as tu



# Test inputs.
_INPUTS = (
    (LIB.NODE_TYPEKEY_AUTHORITY, tu.AUTHORITY_NAME, tu.AUTHORITY_SYNONYMS, None),
    (LIB.NODE_TYPEKEY_SCOPE, tu.SCOPE_NAME, tu.SCOPE_SYNONYMS, None),
    (LIB.NODE_TYPEKEY_COLLECTION, tu.COLLECTION_01_NAME, tu.COLLECTION_01_SYNONYMS, None),
    (LIB.NODE_TYPEKEY_COLLECTION, tu.COLLECTION_02_NAME, tu.COLLECTION_02_SYNONYMS, None),
    (LIB.NODE_TYPEKEY_COLLECTION, tu.COLLECTION_03_NAME, tu.COLLECTION_03_SYNONYMS, None),
    (LIB.NODE_TYPEKEY_TERM, tu.TERM_01_NAME, tu.TERM_01_SYNONYMS, tu.COLLECTION_01_NAME),
    (LIB.NODE_TYPEKEY_TERM, tu.TERM_02_NAME, tu.TERM_02_SYNONYMS, tu.COLLECTION_02_NAME),
    # (LIB.NODE_TYPEKEY_TERM, tu.TERM_03_NAME, tu.TERM_03_SYNONYMS, tu.COLLECTION_03_NAME),
    )


@nose.with_setup(tu.setup, tu.teardown)
def test_parse_name():
    """Test parsing of names at various levels.

    """
    for typekey, canonical_name, alternative_names, parent in _INPUTS:
        for alternative_name in alternative_names:
            for name, expected, strictness in _get_config(canonical_name, alternative_name):
                desc = 'parse --> {}: {} [strictness={}]'.format(typekey, name, strictness)
                tu.init(_test_parse_name, desc)
                yield _test_parse_name, typekey, name, expected, strictness, parent


def _test_parse_name(typekey, name, expected, strictness, parent):
    """Asserts name test.

    """
    if typekey == LIB.NODE_TYPEKEY_AUTHORITY:
        namespace = name
    elif typekey == LIB.NODE_TYPEKEY_SCOPE:
        namespace = '{}:{}'.format(tu.AUTHORITY_NAME, name)
    elif typekey == LIB.NODE_TYPEKEY_COLLECTION:
        namespace = '{}:{}:{}'.format(tu.AUTHORITY_NAME, tu.SCOPE_NAME, name)
    elif typekey == LIB.NODE_TYPEKEY_TERM:
        namespace = '{}:{}:{}:{}'.format(tu.AUTHORITY_NAME, tu.SCOPE_NAME, parent, name)

    try:
        result = LIB.parse(namespace, strictness=strictness)
    except LIB.ParsingError:
        result = LIB.ParsingError

    assert result == expected, \
           'Name parsing error: node-type={}.  name={}.  actual = {}.  expected {}.'.format(typekey, name, result, expected)


def _get_config(canonical_name, alternative_name):
    """Returns node test configuration.

    """
    return [
        (canonical_name, canonical_name, 0),
        # (canonical_name.upper(), LIB.ParsingError, 1),
        # (canonical_name.upper(), LIB.ParsingError, 2),
        # (canonical_name.upper(), canonical_name, 3),
        # (alternative_name, LIB.ParsingError, 0),
        # (alternative_name, LIB.ParsingError, 1),
        # (alternative_name.upper(), LIB.ParsingError, 2),
        # (alternative_name, canonical_name, 2)
    ]




# -*- coding: utf-8 -*-

"""
.. module:: pyessv._factory.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Encapsulates creation of domain model class instances.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>

"""
import arrow

from pyessv._constants import REGEX_CANONICAL_NAME
from pyessv._constants import PARSING_STRICTNESS_2
from pyessv._constants import PARSING_STRICTNESS_SET
from pyessv._cache import cache
from pyessv._exceptions import ValidationError
from pyessv._model import Authority
from pyessv._model import Collection
from pyessv._model import Node
from pyessv._model import Scope
from pyessv._model import Term
from pyessv._parser_template import TemplateParser
from pyessv._utils.compat import basestring
from pyessv._utils.compat import str
from pyessv._utils.formatter import format_canonical_name
from pyessv._utils.formatter import format_string
from pyessv._validation import validate



def create_authority(
    name,
    description,
    label=None,
    url=None,
    create_date=None,
    data=None,
    alternative_names=[]
    ):
    """Instantiates, initialises & returns a term authority.

    :param str name: Canonical name.
    :param str description: Informative description.
    :param str label: Label for UI purposes.
    :param str url: Further information URL.
    :param datetime create_date: Creation date.
    :param dict data: Arbirtrary data.
    :param list alternative_names: Collection of associated alternative names.

    :returns: A vocabulary authority, e.g. wcrp.
    :rtype: pyessv.Authority

    """
    return _create_node(
        typeof=Authority,
        raw_name=name,
        description=description,
        label=label,
        url=url,
        create_date=create_date,
        alternative_names=alternative_names,
        data=data
        )


def create_scope(
    authority,
    name,
    description,
    label=None,
    url=None,
    create_date=None,
    data=None,
    alternative_names=[]
    ):
    """Instantiates, initialises & returns a term scope.

    :param pyessv.Authority authority: CV authority to which scope is bound.
    :param str name: Canonical name.
    :param str description: Informative description.
    :param str label: Label for UI purposes.
    :param str url: Further information URL.
    :param datetime create_date: Creation date.
    :param dict data: Arbirtrary data.
    :param list alternative_names: Collection of associated alternative names.

    :returns: A vocabulary scope, e.g. cmip6.
    :rtype: pyessv.Scope

    """
    def _callback(instance):
        instance.authority = authority
        authority.scopes.append(instance)

    return _create_node(
        typeof=Scope,
        raw_name=name,
        description=description,
        label=label,
        url=url,
        create_date=create_date or authority.create_date,
        alternative_names=alternative_names,
        data=data,
        callback=_callback
        )


def create_collection(
    scope,
    name,
    description,
    label=None,
    url=None,
    create_date=None,
    data=None,
    alternative_names=[],
    term_regex=None
    ):
    """Instantiates, initialises & returns a regular expression term collection.

    :param pyessv.Scope scope: CV scope to which collection is bound.
    :param str name: Canonical name.
    :param str description: Informative description.
    :param str label: Label for UI purposes.
    :param str url: Further information URL.
    :param datetime create_date: Creation date.
    :param dict data: Arbirtrary data.
    :param list alternative_names: Collection of associated alternative names.
    :param str|tuple term_regex: Regular expression information to be applied to terms.

    :returns: A vocabulary collection, e.g. insitution-id.
    :rtype: pyessv.Collection

    """
    if term_regex is None:
        term_regex = REGEX_CANONICAL_NAME

    def _callback(instance):
        instance.scope = scope
        instance.term_regex = term_regex
        scope.collections.append(instance)

    return _create_node(
        typeof=Collection,
        raw_name=name,
        description=description,
        label=label,
        url=url,
        create_date=create_date or scope.create_date,
        alternative_names=alternative_names,
        data=data,
        callback=_callback
        )

def create_term(
    collection,
    name,
    description=None,
    label=None,
    url=None,
    create_date=None,
    data=None,
    alternative_names=[]
    ):
    """Instantiates, initialises & returns a term.

    :param pyessv.Collection collection: The collection to which the term belongs.
    :param str name: Canonical name.
    :param str description: Informative description.
    :param str label: Label for UI purposes.
    :param str url: Further information URL.
    :param datetime create_date: Creation date.
    :param dict data: Arbirtrary data.
    :param list alternative_names: Collection of associated alternative names.

    :returns: A vocabulary term, e.g. ipsl.
    :rtype: pyessv.Term

    """
    def _callback(instance):
        instance.collection = collection
        collection.terms.append(instance)

    return _create_node(
        typeof=Term,
        raw_name=name,
        description=description,
        label=label,
        url=url,
        create_date=create_date or collection.create_date,
        alternative_names=alternative_names,
        data=data,
        callback=_callback
        )


def create_template_parser(template, collections, strictness=PARSING_STRICTNESS_2, seperator='.'):
    """Instantiates, initialises & returns a template parser.

    :param str template: An expression template.
    :param tuple collections: Collections that the template maps to.
    :param int strictness: Strictness level to apply when applying name matching rules.
    :param str seprarator: Seperator to apply when parsing.

    :returns: A vocabulary expression parser.
    :rtype: pyessv.TemplateParser

    """
    assert isinstance(template, basestring), 'Invalid template'
    assert isinstance(collections, tuple), 'Invalid collections'
    assert len(template) > 0, 'Invalid template'
    assert template.count('{}') > 0, 'Invalid template'
    assert len(collections) > 0, 'Invalid collections'
    assert template.count('{}') == len(collections), 'Invalid template: collection count mismatch'
    assert strictness in PARSING_STRICTNESS_SET, 'Invalid parsing strictness: {}'.format(strictness)
    assert isinstance(seperator, basestring), 'Invalid seperator'

    return TemplateParser(template, collections, strictness, seperator)


def _create_node(
    typeof,
    raw_name,
    description,
    label,
    url,
    create_date,
    alternative_names,
    data,
    callback = None
    ):
    """Instantiates, initialises & returns a node.

    """
    # Set core attributes.
    node = typeof()
    node.alternative_names = alternative_names
    node.canonical_name = format_canonical_name(raw_name)
    node.create_date = create_date or arrow.utcnow().datetime
    node.data = data
    node.label = label or format_string(raw_name)
    node.raw_name = format_string(raw_name)

    # Set other attributes.
    if description is not None:
        node.description = format_string(description)
    if url is not None:
        node.url = format_string(url)

    # Invoke node specific callback.
    if callback is not None:
        callback(node)

    # Validate.
    errors = validate(node)
    if errors:
        raise ValidationError(errors)

    # Cache.
    cache(node)

    return node



# -*- coding: utf-8 -*-
"""

.. module:: schemas.extender.py
   :license: GPL/CeCIL
   :platform: Unix
   :synopsis: ES-DOC Errata - endpoint validation schema cache extender.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import collections

from pyessv._ws.utils import constants


# Schema extender functions mapped by schema type and endpoint.
_EXTENDERS = collections.defaultdict(dict)


def extend(schema, typeof, endpoint):
	"""Extends a JSON schema with data pulled from controlled vocabularies.

	:param dict schema: A JSON schema being extended.
	:param str typeof: Type of JSON schema to be extended.
	:param str endpoint: Endpoint being mapped to a JSON schema.

	"""
	try:
		extender = _EXTENDERS[endpoint][typeof]
	except KeyError:
		pass
	else:
		extender(schema)



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._archive.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Encapsulates access to archive.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyessv._cache import cache
from pyessv._cache import get_cached
from pyessv._io_manager import write
from pyessv._model import Authority



def archive(authority):
    """Archive authority to file system.

    """
    cache(authority)
    for authority in get_cached(Authority):
        write(authority)



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._loader.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Orchestrates reading & caching of entities.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import random
import uuid

from pyessv._cache import cache
from pyessv._cache import get_cached
from pyessv._constants import PARSING_NODE_FIELDS
from pyessv._factory import create_term
from pyessv._model import Authority
from pyessv._model import Term
from pyessv._utils import logger
from pyessv._utils.compat import basestring
from pyessv._utils.compat import str
from pyessv._utils.formatter import format_string



def load(identifier=None, verbose=True):
    """Loads a vocabulary node from archive.

    :param str identifier: Vocabulary node identifier.

    :return: A vocabulary node.
    :rtype: pyessv.Node | None

    """
    assert isinstance(identifier, (type(None), basestring))

    if identifier is None:
        return set(get_cached(Authority))

    identifier = identifier.strip()
    result = _load_by_namespace(identifier)
    if result is None and verbose:
        logger.log_warning('Cannot map identifier to a vocabulary entity: {}'.format(identifier))

    return result


def _load_by_namespace(identifier):
    """Loads a vocabulary node from archive by trying to match it's namespace.

    :param str identifier: Vocabulary node namespace.

    :returns: First matching vocabulary node.
    :rtype: pyessv.Node | None

    """
    # Skip if identifier is not a namespace.
    ns = str(identifier).split(':')
    if len(ns) == 0:
        return get_cached(Authority)
    if len(ns) > 4:
        return None

    # Unpack.
    authority = scope = collection = term = None
    if len(ns) == 1:
        authority = ns[0]
    elif len(ns) == 2:
        authority, scope = ns
    elif len(ns) == 3:
        authority, scope, collection = ns
    elif len(ns) == 4:
        authority, scope, collection, term = ns

    # Walk nodes returning deepest match.
    for a in get_cached(Authority):
        if not _is_matched(a, authority):
            continue
        if scope is None:
            return a
        # ... scopes
        for s in a:
            if not _is_matched(s, scope):
                continue
            if collection is None:
                return s
            # ... collections
            for c in s:
                if not _is_matched(c, collection):
                    continue
                if term is None:
                    return c
                # ... terms (concrete)
                for t in c:
                    if _is_matched(t, term):
                        return t
                # ... terms (virtual)
                if c.is_matched(term):
                    return create_term(c, term)


def _is_matched(node, identifier):
    """Returns flag indicating whether node identifier is a match.

    """
    identifier = format_string(identifier).lower()

    # Matched by canonical name.
    if identifier == node.canonical_name:
        return True

    # Matched by raw name.
    if identifier == node.raw_name.lower():
        return True

    # Matched by alternative names.
    elif identifier in [format_string(i).lower() for i in node.alternative_names]:
        return True

    return False


def load_random(namespace, field='canonical_name'):
    """Returns a random term.

    :param str namespace: Namespace of collection from which a term will be loaded.

    :returns: A random term's canonical name.
    :rtype: str

    """
    collection = load(namespace)
    if collection is None:
        raise ValueError('Collection not found: {}'.format(namespace))
    if field not in PARSING_NODE_FIELDS:
        raise ValueError('Invalid field name')

    if collection.is_virtual:
        return str(uuid.uuid4()).split('-')[0]

    term = random.choice(collection.terms)

    return getattr(term, field)



from pyessv._accessors import esdoc_cmip6
from pyessv._accessors import wcrp_cmip6



# Expose accessors mapped by authority.
ACCESSORS = {
	'esdoc': {
		'cmip6': esdoc_cmip6
	},
	'wcrp': {
		'cmip6': wcrp_cmip6
	}
}



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._parsers.cordex_dataset_id.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Encapsulates parsing of a CORDEX dataset identifier.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>

"""
from pyessv._factory import create_template_parser
from pyessv._constants import PARSING_STRICTNESS_1



# Template extracted from esgf ini file (for reference purpose only).
_INI_PATTERN = 'cordex.%(product)s.%(domain)s.%(institute)s.%(driving_model)s.%(experiment)s.%(ensemble)s.%(rcm_name)s.%(rcm_version)s.%(time_frequency)s.%(variable)s'

# Test identifier (for reference purpose only).
_TEST_IDENTIFIER = 'cordex.output.AFR-44.MOHC.MOHC-HadGEM2-ES.rcp60.r12i1p1.hadgem3-ra.v1.mon.areacella'

# Template that identifiers must conform to.
_TEMPLATE = 'cordex.{}.{}.{}.{}.{}.{}.{}.{}.{}.{}'

# Collections injected into template.
_COLLECTIONS = (
    'wcrp:cordex:product',
    'wcrp:cordex:domain',
    'wcrp:cordex:institute',
    'wcrp:cordex:driving-model',
    'wcrp:cordex:experiment',
    'wcrp:cordex:ensemble',
    'wcrp:cordex:rcm-name',
    'wcrp:cordex:rcm-version',
    'wcrp:cordex:time-frequency',
    'wcrp:cordex:variable',
    )

# Instantiated & cached parser instance.
_PARSER = None

def parse(identifier):
    """Parses a CMIP6 dataset identifier.

    """
    # Instantiate parser JIT.
    global _PARSER
    if _PARSER is None:
        _PARSER = create_template_parser(_TEMPLATE, _COLLECTIONS, PARSING_STRICTNESS_1)

    # Strip version suffix.
    if '#' in identifier:
      identifier = identifier.split('#')[0]

    return _PARSER.parse(identifier)



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._model.term.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: A vocabulary term, e.g. IPSL.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyessv._constants import GOVERNANCE_STATUS_PENDING
from pyessv._constants import GOVERNANCE_STATUS_SET
from pyessv._constants import NODE_TYPEKEY_TERM
from pyessv._model.node import Node
from pyessv._utils.compat import basestring
from pyessv._utils.compat import str
from pyessv._utils.validation import assert_regex
from pyessv._utils.validation import assert_string



class Term(Node):
    """A vocabulary term.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(Term, self).__init__(NODE_TYPEKEY_TERM)

        self.associations = list()
        self.collection = None
        self.parent = None
        self.status = GOVERNANCE_STATUS_PENDING


    def __contains__(self, key):
        """Instance membership predicate.

        """
        key = str(key).strip().lower()

        return key in self.all_names


    @property
    def ancestors(self):
        """Gets ancestors within archive hierarchy.

        """
        return [self.authority, self.scope, self.collection]


    @property
    def authority(self):
        """Gets governing authority.

        """
        return self.scope.authority


    @property
    def scope(self):
        """Gets governing scope.

        """
        return self.collection.scope


    def get_validators(self):
        """Returns set of validators.

        """
        from pyessv._model.scope import Collection

        def _canonical_name():
            assert_string(self.canonical_name)
            if isinstance(self.collection.term_regex, basestring):
                assert_regex(self.canonical_name, self.collection.term_regex)

        def _collection():
            assert isinstance(self.collection, Collection)

        def _parent():
            if self.parent is not None:
                assert isinstance(self.parent, Term)

        def _status():
            assert self.status in GOVERNANCE_STATUS_SET

        return super(Term, self).get_validators() + (
            _collection,
            _canonical_name,
            _parent,
            _status,
            )



# -*- coding: utf-8 -*-

"""
.. module:: test_model.py

   :copyright: @2013 Earth System Documentation (https://es-doc.org)
   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Executes pyessv model tests.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import inspect
import nose

import pyessv as LIB
import tests.utils as tu



# Test configuration: project, parsing function, template seperator, strictness, identifiers.
_CONFIG = {
    ('cmip5', LIB.parse_dataset_identifer, '.', (
        'cmip5.output1.IPSL.IPSL-CM5A-LR.aqua4K.3hr.atmos.3hr.r2i1p1',
        'cmip5.output2.IPSL.IPSL-CM5A-LR.historicalMisc.mon.ocean.Omon.r2i1p1'
    )),
    ('cmip6', LIB.parse_dataset_identifer, '.', (
        'cmip6.FAFMIP.IPSL.IPSL-CM6A-LR.amip.r1i1p1f1.Amon.abs550aer.gm',
    )),
    ('cordex', LIB.parse_dataset_identifer, '.', (
        'cordex.output.AFR-44.MOHC.MOHC-HadGEM2-ES.rcp60.r12i1p1.HadGEM3-RA.v1.mon.areacella',
        'cordex.output.EUR-11.SMHI.ICHEC-EC-EARTH.rcp85.r12i1p1.RCA4.v1.sem.rsdt'
    )),
}



def test_parse_identifiers():
    """pyessv-tests: parsing: identifiers

    """
    def positive_test(parser, project, identifier):
        parser(project, identifier)

    @nose.tools.raises(LIB.TemplateParsingError)
    def negative_test(parser, project, identifier):
        parser(project, identifier)

    # Iterate identifiers & perform +ve / -ve tests:
    for project, parser, seperator, identifiers in _CONFIG:
        assert inspect.isfunction(parser)
        for identifier in identifiers:
            # ... +ve test:
            desc = 'identifier parsing test (+ve) --> {} :: {}'.format(project, identifier)
            tu.init(positive_test, desc)
            yield positive_test, parser, project, identifier

            # ... -ve tests:
            for invalid_identifier in _get_invalid_identifiers(identifier, seperator):
                desc = 'identifier parsing test (-ve) --> {} :: {}'.format(project, invalid_identifier)
                tu.init(negative_test, desc)
                yield negative_test, parser, project, invalid_identifier


def _get_invalid_identifiers(identifier, seperator):
    """Returns a set of invalid identifiers by fuxxing a valid identifier.

    """
    # Set parts.
    parts = identifier.split(seperator)

    # Invalid as the number of parts is not as expected.
    yield seperator.join(parts[1:])

    # Invalid as a part has been replaced with invalid value.
    for idx in range(len(parts)):
        new_parts = list(parts)
        new_parts[idx] = 'X!X!X!X!'
        yield seperator.join(new_parts)



# -*- coding: utf-8 -*-
"""

.. module:: schemas.cache.py
   :license: GPL/CeCIL
   :platform: Unix
   :synopsis: ES-DOC Errata - endpoint validation schema cache.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import collections

from pyessv._ws.schemas import loader



# Cached store of loaded schemas.
_STORE = collections.defaultdict(dict)

# Set of endpoint substitutions.
_SUBSTITUTIONS = {
	'/1/retrieve',
}


def init(endpoints):
	"""Initializes cache from schemas upon file system.

	:param dict endpoints: Map of application endpoints.

	"""
    # Load retrieve json scehmas.
	for typeof in {'body', 'params', 'headers'}:
		_STORE[typeof]['/1/retrieve'] = loader.load(typeof, '/1/retrieve')

    # Load other json scehmas.
	for endpoint in [i for i in endpoints if not i.startswith('/1/retrieve')]:
		for typeof in {'body', 'params', 'headers'}:
            # payload = loader.load(typeof, endpoint)
			_STORE[typeof][endpoint] = loader.load(typeof, endpoint)


def get_schema(typeof, endpoint):
	"""Gets a schema from cache.

	"""
	for ep in _SUBSTITUTIONS:
		if endpoint.startswith(ep):
			endpoint = ep

	try:
		return _STORE[typeof][endpoint]
	except KeyError:
		pass



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._model.__init__.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Model of vocabulary entities.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyessv._model.authority import Authority
from pyessv._model.scope import Scope
from pyessv._model.collection import Collection
from pyessv._model.node import Node
from pyessv._model.node import IterableNode
from pyessv._model.term import Term

# Set of node types.
NODE_TYPES = {
    Authority,
    Collection,
    Scope,
    Term
}



# -*- coding: utf-8 -*-

"""
.. module:: test_io_manager.py

   :copyright: @2013 Earth System Documentation (https://es-doc.org)
   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Executes pyessv I/O tests.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import inspect
import io
import json
import nose
import os

import pyessv as LIB
from pyessv._io_manager import delete
from pyessv._io_manager import read
from pyessv._io_manager import write
import tests.utils as tu



def test_interface():
    """pyessv-tests: io: interface.

    """
    assert inspect.isfunction(delete)
    assert inspect.isfunction(read)
    assert inspect.isfunction(write)


def test_directory():
    """pyessv-tests: io: directory.

    """
    assert os.path.isdir(LIB.DIR_ARCHIVE)


def test_read():
    """pyessv-tests: io: read.

    """
    authorities = read()
    assert isinstance(authorities, list)
    for authority in authorities:
        assert isinstance(authority, LIB.Authority)

    dirs = os.listdir(LIB.DIR_ARCHIVE)
    assert len(authorities) == len(dirs)
    assert dirs == [i.canonical_name for i in authorities]


@nose.with_setup(tu.setup, tu.teardown)
def test_write():
    """pyessv-tests: io: write.

    """
    authority_dirs = os.listdir(LIB.DIR_ARCHIVE)
    authority_dir = os.path.join(LIB.DIR_ARCHIVE, tu.AUTHORITY_NAME)
    authority_manifest = os.path.join(authority_dir, 'MANIFEST')
    scope_dir = os.path.join(authority_dir, tu.SCOPE_NAME)
    collection_01_dir = os.path.join(scope_dir, tu.COLLECTION_01_NAME)
    collection_02_dir = os.path.join(scope_dir, tu.COLLECTION_02_NAME)
    collection_03_dir = os.path.join(scope_dir, tu.COLLECTION_03_NAME)
    term_01_file = os.path.join(collection_01_dir, tu.TERM_01_NAME)
    term_02_file = os.path.join(collection_02_dir, tu.TERM_02_NAME)
    term_03_file = os.path.join(collection_03_dir, tu.TERM_03_NAME)

    dpaths = (authority_dir, scope_dir, collection_01_dir, collection_02_dir, collection_03_dir)
    fpaths = (authority_manifest, term_01_file, term_02_file, term_03_file)

    for dpath in dpaths:
        assert not os.path.isdir(dpath)
    for fpath in fpaths:
        assert not os.path.isfile(fpath)

    write(LIB.load(tu.AUTHORITY_NAME))

    assert len(os.listdir(LIB.DIR_ARCHIVE)) == len(authority_dirs) + 1
    for dpath in dpaths:
        assert os.path.isdir(dpath)
    for fpath in fpaths:
        assert os.path.isfile(fpath)

    with io.open(authority_manifest, 'r') as fstream:
        assert isinstance(json.loads(fstream.read()), dict)
    for fpath in fpaths:
        with io.open(fpath, 'r') as fstream:
            assert isinstance(json.loads(fstream.read()), dict)


@nose.with_setup(tu.setup, tu.teardown)
def test_delete():
    """pyessv-tests: io: delete.

    """
    authority_dirs = os.listdir(LIB.DIR_ARCHIVE)
    authority_dir = os.path.join(LIB.DIR_ARCHIVE, tu.AUTHORITY_NAME)
    authority_manifest = os.path.join(authority_dir, 'MANIFEST')
    scope_dir = os.path.join(authority_dir, tu.SCOPE_NAME)
    collection_01_dir = os.path.join(scope_dir, tu.COLLECTION_01_NAME)
    collection_02_dir = os.path.join(scope_dir, tu.COLLECTION_02_NAME)
    collection_03_dir = os.path.join(scope_dir, tu.COLLECTION_03_NAME)
    term_01_file = os.path.join(collection_01_dir, tu.TERM_01_NAME)
    term_02_file = os.path.join(collection_02_dir, tu.TERM_02_NAME)
    term_03_file = os.path.join(collection_03_dir, tu.TERM_03_NAME)
    write(LIB.load(tu.AUTHORITY_NAMESPACE))

    for namespace, npath, predicate in (
        (tu.TERM_01_NAMESPACE, term_01_file, os.path.isfile),
        (tu.TERM_02_NAMESPACE, term_02_file, os.path.isfile),
        (tu.TERM_03_NAMESPACE, term_03_file, os.path.isfile),
        (tu.COLLECTION_01_NAMESPACE, collection_01_dir, os.path.isdir),
        (tu.COLLECTION_02_NAMESPACE, collection_02_dir, os.path.isdir),
        (tu.COLLECTION_03_NAMESPACE, collection_03_dir, os.path.isdir),
        (tu.SCOPE_NAMESPACE, scope_dir, os.path.isdir),
        (tu.AUTHORITY_NAMESPACE, authority_dir, os.path.isdir),
        ):
        node = LIB.load(namespace)
        delete(node)
        assert not predicate(npath)



# -*- coding: utf-8 -*-
"""

.. module:: app.py
   :license: GPL/CeCIL
   :platform: Unix
   :synopsis: ES-DOC Errata - web-service entry point.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import os

import tornado.web

import pyessv
from pyessv._ws import handlers
from pyessv._ws import schemas
from pyessv._ws.utils import config
from pyessv._ws.utils.logger import log_web as log



def _get_path_to_front_end():
    """Return path to the front end javascript application.

    """
    dpath = os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(__file__))), 'fe')
    log("Front-end static files @ {0}".format(dpath))

    return dpath


def _get_app_endpoints():
    """Returns map of application endpoints to handlers.

    """
    endpoints = [r'/1/retrieve']
    for authority in pyessv.get_cached():
        endpoints.append(r'/1/retrieve/{}'.format(authority.canonical_name))
        for scope in authority:
            endpoints.append(r'/1/retrieve/{}/{}'.format(
                authority.canonical_name,
                scope.canonical_name)
                )
            for collection in scope:
                endpoints.append(r'/1/retrieve/{}/{}/{}'.format(
                    authority.canonical_name,
                    scope.canonical_name,
                    collection.canonical_name)
                    )
                for term in collection:
                    endpoints.append(r'/1/retrieve/{}/{}/{}/{}'.format(
                        authority.canonical_name,
                        scope.canonical_name,
                        collection.canonical_name,
                        term.canonical_name)
                        )
    endpoints = [(i, handlers.RetrieveRequestHandler) for i in endpoints]
    endpoints += [
        (r'/', handlers.HeartbeatRequestHandler),
        (r'/status', handlers.HeartbeatRequestHandler),
        (r'/1/validate-identifier', handlers.ValidateIdentifierHandler),
        (r'/1/validate-identifier-set', handlers.ValidateIdentifierSetHandler)
        ]

    return set(endpoints)


def _get_app_settings():
    """Returns app settings.

    """
    return {
        "cookie_secret": config.cookie_secret,
        "compress_response": True
    }


def _get_app():
    """Returns application instance.

    """
    endpoints = _get_app_endpoints()
    log("Assigned endpoint to handler mappings")

    schemas.init([i[0] for i in endpoints])
    return tornado.web.Application(endpoints,
                                   debug=True,
                                   **_get_app_settings())


def run():
    """Runs web service.

    """
    # Initialize application.
    log("Initializing")
    app = _get_app()

    # Open port.
    app.listen(config.port)
    log("Ready")

    # Start processing requests.
    tornado.ioloop.IOLoop.instance().start()


def stop():
    """Stops web service.

    """
    ioloop = tornado.ioloop.IOLoop.instance()
    ioloop.add_callback(lambda x: x.stop(), ioloop)



# -*- coding: utf-8 -*-

"""
.. module:: utils.exceptions.py
   :platform: Unix
   :synopsis: Custom exceptions used in this module for better readability of code.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyessv._ws.utils import constants



class WebServiceError(Exception):
    """Web service error wrapper.

    """
    def __init__(self, msg, response_code):
        """Instance constructor.

        """
        super(WebServiceError, self).__init__(msg)
        self.response_code = response_code


class RequestValidationException(WebServiceError):
    """Base class for request validation exceptions.

    """
    def __init__(self, msg):
        """Instance constructor.

        """
        super(RequestValidationException, self).__init__(msg, constants.HTTP_RESPONSE_BAD_REQUEST_ERROR)


class InvalidJSONError(RequestValidationException):
    """Raised if the submitted request data is invalid according to a JSON schema.

    """
    def __init__(self, json_err):
        """Instance constructor.

        """
        msg = json_err.message.strip()
        try:
            self.field = json_err.path[0]
        except Exception as err:
            self.field = msg.split("'")[1]
        super(InvalidJSONError, self).__init__(msg)


class UnknownEndpointException(WebServiceError):
    """Base class for unknown endpoint exceptions.

    """
    def __init__(self, msg):
        """Instance constructor.

        """
        super(UnknownEndpointException, self).__init__(msg, constants.HTTP_RESPONSE_BAD_REQUEST_ERROR)


class ConfigFileNotFoundException(WebServiceError):
    """Configuration file not found exception.

    """
    def __init__(self, path):
        """Instance constructor.

        """
        msg = "configuration file could not be found : {}".format(path)
        super(ConfigFileNotFoundException, self).__init__(msg, constants.HTTP_RESPONSE_SERVER_ERROR)


# Map of managed error codes.
ERROR_CODES = {
    InvalidJSONError: 900,
    ConfigFileNotFoundException: 901,
}



# -*- coding: utf-8 -*-
"""
.. module:: utils.http_validator.py
   :license: GPL/CeCIL
   :platform: Unix
   :synopsis: HTTP request validation.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import json

import jsonschema

from pyessv._ws.utils import exceptions
from pyessv._ws.schemas import get_schema


DDD = {
    '/1/retrieve',
}


def validate_request(handler):
    """Validates request against mapped JSON schemas.

    :param utils.http.HTTPRequestHandler handler: An HTTP request handler.

    :raises: exceptions.InvalidJSONError

    """
    for func in {
        _validate_request_headers,
        _validate_request_params,
        _validate_request_body
        }:
        func(handler)


def _validate_request_headers(handler):
    """Validates request headers against a JSON schema.

    """
    # Map request to schema.
    schema = get_schema('headers', handler.request.path)

    # Null case - escape.
    if schema is None:
        return

    # Validate request headers.
    _validate(handler, dict(handler.request.headers), schema)


def _validate_request_params(handler):
    """Validates request parameters against a JSON schema.

    """
    # Map request to schema.
    schema = get_schema('params', handler.request.path)

    # Null case.
    if schema is None:
        if handler.request.query_arguments:
            raise exceptions.RequestValidationException("Unexpected request url parameters.")

    # Validate request parameters.
    else:
        _validate(handler, handler.request.query_arguments, schema)


def _validate_request_body(handler):
    """Validates request body against a JSON schema.

    """
    # Map request to schema.
    schema = get_schema('body', handler.request.path)

    # Null case.
    if schema is None:
        if handler.request.body:
            raise exceptions.RequestValidationException("Unexpected request body.")

    # Validate request data.
    else:
        # ... decode request data.
        data = json.loads(handler.request.body)

        # ... validate request data against schema.
        _validate(handler, data, schema)

        # ... append valid data to request.
        handler.request.data = data


def _validate(handler, data, schema):
    """Validates data against a JSON schema.

    """
    try:
        jsonschema.validate(data, schema)
    except jsonschema.exceptions.ValidationError as err:
        raise exceptions.InvalidJSONError(err)



# -*- coding: utf-8 -*-
"""
.. module:: utils.factory.py
   :license: GPL/CeCIL
   :platform: Unix
   :synopsis: Test issue factory.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import pyessv


# Collection of materials, i.e. supporting images, graphs ... etc.
_MATERIALS = []

# Collection of data set patters used when generating test dataset identifiers.
_DATASETS_PATTERNS = {
    'cmip5': u'cmip5.{}.{}.{}.{}.{}.{}.{}.r1i1p1#v20180101',
    'cmip6': u'CMIP6.{}.{}.{}.{}.r1i1p1f1.{}.Emon.{}#v20180101',
    'cordex': u'cordex.{}.{}.{}.{}.{}.r12i1p1.{}.v1.{}.{}#v20180101'
}


def get_test_datasets(project, existing=[]):
    """Returns a collection of test dataset identifiers.

    :param str project: Project code.
    :param list existing: Dataset identifiers to be included in the result.

    :returns: Collection of test datasets.
    :rtype: list

    """
    return [_get_test_dataset(project) for i in range(5)] + existing


def _get_test_dataset(project):
    """Returns a test dataset identifier.

    """
    pattern = _DATASETS_PATTERNS[project]

    if project == 'cmip5':
        return pattern.format(
            pyessv.load_random('wcrp:cmip5:product', field='raw_name'),
            pyessv.parse('wcrp:cmip5:institute:ipsl', field='raw_name'),
            pyessv.load_random('wcrp:cmip5:model', field='raw_name'),
            pyessv.load_random('wcrp:cmip5:experiment', field='raw_name'),
            pyessv.load_random('wcrp:cmip5:time-frequency', field='raw_name'),
            pyessv.load_random('wcrp:cmip5:realm', field='raw_name'),
            pyessv.load_random('wcrp:cmip5:cmor-table', field='raw_name')
            )

    elif project == 'cmip6':
        return pattern.format(
            pyessv.load_random('wcrp:cmip6:activity-id', field='raw_name'),
            pyessv.parse('wcrp:cmip6:institution-id:ipsl', field='raw_name'),
            pyessv.load_random('wcrp:cmip6:source-id', field='raw_name'),
            pyessv.load_random('wcrp:cmip6:experiment-id', field='raw_name'),
            pyessv.load_random('wcrp:cmip6:table-id', field='raw_name'),
            pyessv.load_random('wcrp:cmip6:grid-label', field='raw_name')
            )

    elif project == 'cordex':
        return pattern.format(
            pyessv.load_random('wcrp:cordex:product', field='raw_name'),
            pyessv.load_random('wcrp:cordex:domain', field='raw_name'),
            pyessv.parse('wcrp:cordex:institute:ipsl-ineris', field='raw_name'),
            pyessv.load_random('wcrp:cordex:driving-model', field='raw_name'),
            pyessv.load_random('wcrp:cordex:experiment', field='raw_name'),
            pyessv.load_random('wcrp:cordex:rcm-name', field='raw_name'),
            pyessv.load_random('wcrp:cordex:time-frequency', field='raw_name'),
            pyessv.load_random('wcrp:cordex:variable', field='raw_name')
            )



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._initializer.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Initializes library.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import inspect
import os

import pyessv
from pyessv._accessors import ACCESSORS
from pyessv._cache import cache
from pyessv._constants import DIR_ARCHIVE
from pyessv._io_manager import read
from pyessv._utils import logger



def init():
	"""Library initializer.

	"""
	# Verify archive folder exists.
	if not os.path.isdir(DIR_ARCHIVE):
		raise EnvironmentError('{} directory does not exists'.format(DIR_ARCHIVE))

	# Load set of authorities from file system.
	authorities = _load_authorities()

	# Mixin pseudo-constants.
	_mixin_constants(authorities)

	# Set scope level accessor functions.
	_mixin_scope_accessors(authorities)


def _load_authorities():
	"""Loads vocabulary authorities from archive.

	"""
	logger.log('Loading vocabularies from {}:'.format(DIR_ARCHIVE))
	authorities = []
	for authority in read():
		authorities.append(authority)
		logger.log('... loaded: {}'.format(authority))
		cache(authority)

	return authorities


def _mixin_constants(authorities):
	"""Mixes in authorities as pseudo-constants to pyessv.

	"""
	for authority in authorities:
		attr_name = authority.canonical_name.replace('-', '_').upper()
		setattr(pyessv, attr_name, authority)


def _mixin_scope_accessors(authorities):
	"""Mixes in scope level vocab accessors functions.

	"""
	# In pyessv._accessors sub-package are modules that expose helper functions for accessing vocabularies,
	# here we are ensuring that those functions are easily accessed.
	targets = []
	for authority in authorities:
		for scope in authority:
			try:
				accessor = ACCESSORS[authority.canonical_name][scope.canonical_name]
			except KeyError:
				pass
			else:
				targets.append((scope, accessor))

	# Mixin accessor functions with scope.
	for scope, accessor in targets:
		funcs = [i for i in inspect.getmembers(accessor)
				 if inspect.isfunction(i[1]) and not i[0].startswith('_')]
		for name, func in funcs:
			setattr(scope, name, func)



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._parsers.__init__.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Expression parsers.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import collections

from pyessv._parsers.cmip5_dataset_id import parse as parse_cmip5_dataset_id
from pyessv._parsers.cmip6_dataset_id import parse as parse_cmip6_dataset_id
from pyessv._parsers.cordex_dataset_id import parse as parse_cordex_dataset_id
from pyessv._parsers.input4mips_dataset_id import parse as parse_input4mips_dataset_id
from pyessv._utils.compat import basestring



# Map of dataset id parsers to projects.
_DATASET_ID_PARSERS = {
    'cmip5': parse_cmip5_dataset_id,
    'cmip6': parse_cmip6_dataset_id,
    'cordex': parse_cordex_dataset_id,
    'input4mips': parse_input4mips_dataset_id
}


def parse_dataset_identifers(project, identifiers):
    """Parses a collection of dataset identifiers.

    :param str project: Project code.
    :param iterable identifiers: Dataset identifiers.

    :returns: Facets extracted from the identifiers.
    :rtype: list

    """
    assert isinstance(identifiers, collections.Iterable), 'Invalid identifiers'

    result = set()
    for identifier in identifiers:
        result = result.union(parse_dataset_identifer(project, identifier))

    return result


def parse_dataset_identifer(project, identifier):
    """Parses a dataset identifier.

    :param str project: Project code.
    :param str identifier: Dataset identifier.

    :returns: Set of terms extracted from the identifier.
    :rtype: set

    """
    assert isinstance(project, basestring), 'Invalid project'
    assert project in _DATASET_ID_PARSERS, 'Unsupported project'
    assert isinstance(identifier, basestring), 'Invalid identifier'

    parser = _DATASET_ID_PARSERS[project]

    return parser(identifier)



# -*- coding: utf-8 -*-

"""
.. module:: test_codecs.py

   :copyright: @2013 Earth System Documentation (https://es-doc.org)
   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Executes pyessv encoding / decoding tests.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import inspect
import nose

import pyessv as LIB
from pyessv._codecs import decode
from pyessv._codecs import encode
from pyessv._constants import ENCODING_SET
from pyessv._constants import STANDARD_NODE_FIELDS
from pyessv._utils.compat import basestring
from pyessv._utils.compat import str
import tests.utils as tu



# Types of representation by encoding.
_ENCODING_REPRESENTATION_TYPE = {
    LIB.ENCODING_DICT: dict,
    LIB.ENCODING_JSON: basestring,
}



def test_interface():
    """pyessv-tests: codecs: interface.

    """
    assert inspect.isfunction(decode)
    assert inspect.isfunction(encode)


@nose.with_setup(None, tu.teardown)
def test_encode():
    """pyessv-tests: encode.

    """
    def _test(node, encoding):
        """Inner test.

        """
        representation = encode(node, encoding)
        assert isinstance(representation, _ENCODING_REPRESENTATION_TYPE[encoding])
        decoded = decode(representation, encoding)
        assert isinstance(decoded, type(node))
        for field in STANDARD_NODE_FIELDS:
            assert getattr(decoded, field) == getattr(node, field)


    for node_factory in (
        tu.create_authority,
        tu.create_scope,
        tu.create_collection_01,
        tu.create_collection_02,
        tu.create_collection_03,
        tu.create_term_01,
        tu.create_term_02,
        tu.create_term_03
        ):
        node = node_factory()
        for encoding in ENCODING_SET:
            desc = 'codecs: {} --> {}'.format(node_factory.__name__[7:], encoding)
            tu.init(_test, desc)
            yield _test, node, encoding



# -*- coding: utf-8 -*-
"""
.. module:: utils.config.py
   :license: GPL/CeCIL
   :platform: Unix
   :synopsis: Configuration utility functions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import os
import logger
import json

from pyessv._ws.utils.constants import HTTP_RESPONSE_SERVER_ERROR
from pyessv._ws.utils.convertor import json_file_to_namedtuple
from pyessv._ws.utils.convertor import to_namedtuple
from pyessv._ws.utils.exceptions import ConfigFileNotFoundException



# Default configuration file path.
_CONFIG_FPATH = "ws.conf"

# Default configuration.
_DEFAULT = {
    'cookie_secret': 'p2FAdrUN3tac',
    'host': 'localhost',
    'apply_security_policy': False,
    'mode': 'dev',
    'port': 5010
}

# Configuration data.
data = to_namedtuple(_DEFAULT)


def _get_config_fpath(config_path):
    """Returns configuration file path.

    """
    path = os.getenv('PYESSV_LIB_HOME')
    if path is not None:
        path = os.path.join(path, "ops/config")
        path = os.path.join(path, _CONFIG_FPATH)
        if os.path.exists(path):
            return path


def _init():
    """Initializes configuration.

    """
    global data

    fpath = _get_config_fpath('ws.conf')
    if fpath is not None:
        data = json_file_to_namedtuple(fpath)
        logger.log_web("Config file loaded @ {}".format(fpath))
    else:
        logger.log_web_warning("Web-service config file not found: reverted to default")


# Auto-initialize.
_init()



# -*- coding: utf-8 -*-

"""
.. module:: handlers.validate_identifier.py
   :license: GPL/CeCIL
   :platform: Unix
   :synopsis: ES-DOC pyessv - identifier validation endpoint.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import datetime as dt

import tornado

import pyessv
from pyessv._ws.utils.http import process_request



# Query parameter names.
_PARAM_IDENTIFIER_TYPE = 'identifierType'
_PARAM_PROJECT = 'project'

# Payload body field names.
_FIELD_IDENTIFIER_SET = 'identifierSet'

# Map of identifier type to parser.
_PARSERS = {
    'dataset': pyessv.parse_dataset_identifers,
}


class ValidateIdentifierSetHandler(tornado.web.RequestHandler):
    """Dataset identifier set validation endpoint request handler.

    """
    def post(self):
        """HTTP POST handler.

        """
        def _parse_identifier():
            """Parses an identifier.

            """
            # Decode.
            identifier_type = self.get_argument(_PARAM_IDENTIFIER_TYPE)
            identifiers = self.request.data[_FIELD_IDENTIFIER_SET]
            project = self.get_argument(_PARAM_PROJECT)

            # Parse.
            parser = _PARSERS[identifier_type]
            parser(project, identifiers)


        def _set_output():
            """Sets response to be returned to client.

            """
            self.output = {
                "message": "Datasets are valid {} identifiers".format(
                    self.get_argument(_PARAM_IDENTIFIER_TYPE)
                ),
            }

        # Process request.
        process_request(self, [
            _parse_identifier,
            _set_output
            ])



# -*- coding: utf-8 -*-
"""
.. module:: utils.logger.py
   :license: GPL/CeCIL
   :platform: Unix
   :synopsis: Logging utility functions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import arrow



# Set of logging levels.
LOG_LEVEL_DEBUG = 'DUBUG'
LOG_LEVEL_INFO = 'INFO'
LOG_LEVEL_WARNING = 'WARN'
LOG_LEVEL_ERROR = 'ERROR'
LOG_LEVEL_CRITICAL = 'CRITICAL'
LOG_LEVEL_FATAL = 'FATAL'
LOG_LEVEL_SECURITY = 'SECURITY'

# Defaults.
_DEFAULT_MODULE = "**"

# Text to display when passed a null message.
_NULL_MSG = "-------------------------------------------------------------------------------"


def _get_formatted_message(msg, module, level):
    """Returns a message formatted for logging.

    """
    if msg is None:
        return _NULL_MSG

    return "{} [{}] :: ESDOC-PYESSV-{} :: {}".format(
        unicode(arrow.get())[0:-6],
        level,
        module,
        unicode(msg).strip()
        )


def log(msg=None, module=_DEFAULT_MODULE, level=LOG_LEVEL_INFO):
    """Outputs a message to log.

    :param str msg: Message to be written to log.
    :param str module: Module emitting log message (e.g. MQ).
    :param str level: Message level (e.g. INFO).

    """
    # TODO use structlog rather than printing to stdout
    print(_get_formatted_message(msg, module, level))


def log_warning(msg, module=_DEFAULT_MODULE):
    """Logs a warning event.

    :param str msg: A log message.
    :param str level: Message level (e.g. INFO).

    """
    log(msg, module, LOG_LEVEL_WARNING)


def warn(msg, module=_DEFAULT_MODULE):
    """Logs a warning event.

    :param str msg: A log message.
    :param str level: Message level (e.g. INFO).

    """
    log_warning(msg, module)


def log_error(err, module=_DEFAULT_MODULE):
    """Logs a runtime error.

    :param Exception err: Error to be written to log.
    :param str module: Module emitting log message (e.g. DB).

    """
    msg = "!! {0} RUNTIME ERROR !! :: ".format(module)
    if issubclass(BaseException, err.__class__):
        msg += "{} :: ".format(err.__class__)
    msg += "{}".format(err)
    log(msg, module, LOG_LEVEL_ERROR)


def log_db(msg, level=LOG_LEVEL_INFO):
    """Logs database related events.

    :param str msg: Database message for writing to log.
    :param str level: Message level (e.g. INFO).

    """
    log(msg, "DB", level)


def log_db_warning(msg):
    """Logs a database warning event.

    :param str msg: A log message.
    :param str level: Message level (e.g. INFO).

    """
    log_db(msg, level=LOG_LEVEL_WARNING)


def log_db_error(err):
    """Logs a database error event.

    :param Exception err: Exception to be logged.

    """
    log_error(err, "DB")


def log_pid(msg, level=LOG_LEVEL_INFO):
    """Logs PID service related events.

    :param str msg: PID service message for writing to log.
    :param str level: Message level (e.g. INFO).

    """
    log(msg, "PID", level)


def log_pid_warning(msg):
    """Logs a PID service warning event.

    :param str msg: A log message.
    :param str level: Message level (e.g. INFO).

    """
    log_pid(msg, level=LOG_LEVEL_WARNING)


def log_pid_error(err):
    """Logs a PID service error event.

    :param Exception err: Exception to be logged.

    """
    log_error(err, "PID")


def log_web(msg, level=LOG_LEVEL_INFO):
    """Logs web related events.

    :param str msg: A log message.
    :param str level: Message level (e.g. INFO).

    """
    log(msg, "WS", level)


def log_web_warning(msg):
    """Logs a web warning event.

    :param str msg: A log message.

    """
    log_web(msg, LOG_LEVEL_WARNING)


def log_web_security(msg):
    """Logs a web security event.

    :param str msg: A log message.

    """
    log_web(msg, LOG_LEVEL_SECURITY)


def log_web_error(err):
    """Logs a web error event.

    :param Exception err: Exception to be logged.

    """
    log_error(err, "WS")



# -*- coding: utf-8 -*-

"""
.. module:: test_validation.py

   :copyright: @2013 Earth System Documentation (https://es-doc.org)
   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Executes pyessv validation tests.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import arrow
import nose

import pyessv as LIB
import tests.utils as tu
import tests.utils_model as tum



# Node level test information.
_NODE_TEST_INFO = [
    ('create_date', ('', '  ', 123)),
    ('data', ('', '  ', 123)),
    ('description', ('', '  ')),
    ('canonical_name', (None, '', '  ', 'invalid-CANONICAL-name!')),
    ('url', ('', '  ', 'an-invalid-url')),
    ]


# Test information mapped by node type.
_TEST_INFO = {
    LIB.Authority: _NODE_TEST_INFO + [
        ('scopes', [None, '', '  ', [123]])
    ],
    LIB.Scope: _NODE_TEST_INFO + [
        ('collections', [None, '', '  ', [123]])
    ],
    LIB.Collection: _NODE_TEST_INFO + [
        ('terms', [None, '', '  ', [123]])
    ],
    LIB.Term: _NODE_TEST_INFO + [
        ('status', ('', '  ', [123])),
        ('alternative_names', ('', '  ', [123])),
    ],
}


@nose.with_setup(None, tu.teardown)
def test_node():
    """Tests node validation.

    """
    for node_factory in (
        tu.create_authority,
        tu.create_scope,
        tu.create_collection_01,
        tu.create_collection_02,
        tu.create_collection_03,
        tu.create_term_01,
        tu.create_term_02,
        tu.create_term_03
        ):
        node = node_factory()
        for attr, invalid in _TEST_INFO[type(node)]:
            tu.init(_test_node_attr, 'validate --> {}: {}'.format(node_factory.__name__[7:], attr))
            yield _test_node_attr, node, attr, invalid


def _test_node_attr(instance, attr, invalid_values):
    """Tests node attribute validation.

    """
    valid_value = getattr(instance, attr)
    for value in invalid_values:
        setattr(instance, attr, value)
        assert LIB.is_valid(instance) == False, (attr, value)
        assert len(LIB.get_errors(instance)) >= 1, (LIB.get_errors(instance), attr, value)

    setattr(instance, attr, valid_value)
    assert LIB.is_valid(instance) == True, (LIB.get_errors(instance), attr, valid_value)
    assert len(LIB.get_errors(instance)) == 0


def test_regex_collection():
    """pyessv-tests: validate --> reg-ex collection

    """
    collection = tu.create_collection_01()
    collection.term_regex = r'^[a-z\-]*$'
    term = tu.create_term_01(collection=collection)
    term.canonical_name = 'abc-def'
    assert LIB.is_valid(term) == True


def test_regex_collection_negative():
    """pyessv-tests: validate --> reg-ex collection --> negative

    """
    collection = tu.create_collection_01()
    collection.term_regex = r'^[a-z\-]*$'
    term = tu.create_term_01(collection=collection)
    term.canonical_name = 'ABC-DEF'
    assert LIB.is_valid(term) == False



# -*- coding: utf-8 -*-

"""
.. module:: test_authoring.py

   :copyright: @2013 Earth System Documentation (https://es-doc.org)
   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Executes pyessv authoring tests.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import nose
import pyessv as LIB

from pyessv import load
from pyessv._utils.compat import str
import tests.utils as tu



@nose.with_setup(None, tu.teardown)
def test_create():
    """Test creating an authority.

    """
    def _test(func, typeof):
        """Inner test.

        """
        node = func()
        tu.assert_object(node, typeof)
        assert node == load(node.namespace)


    for factory, typeof in (
        (tu.create_authority, LIB.Authority),
        (tu.create_scope, LIB.Scope),
        (tu.create_collection_01, LIB.Collection),
        (tu.create_collection_02, LIB.Collection),
        (tu.create_collection_03, LIB.Collection),
        (tu.create_term_01, LIB.Term),
        (tu.create_term_02, LIB.Term),
        (tu.create_term_03, LIB.Term)
        ):
        tu.init(_test, 'create --> {}'.format(factory.__name__[7:]))
        yield _test, factory, typeof



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._cache.store.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Manages cache stores.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyessv._cache import store_memory as _memory_store
from pyessv._constants import CACHE_STORE_MEMORY
from pyessv._constants import CACHE_STORE_TYPES
from pyessv._model import Authority
from pyessv._model import Node



# Cache stores.
_STORES = {
    CACHE_STORE_MEMORY: _memory_store
}


def cache(node):
    """Caches a vocabulary node.

    :param pyeesv.Node: Node to be cached.

    """
    assert isinstance(node, Node), 'Invalid node'

    for store in _STORES.values():
        store.cache(node)


def get_cached(identifier=None, store_type=CACHE_STORE_MEMORY):
    """Returns a cached node.

    :param str|None|class filter: Cache filter expression.
    :param str store_type: Cache store type.

    :returns: A pointer to a vocabulary node or a set of nodes.
    :rtype: pyessv.Node | list[Authority]

    """
    assert store_type in CACHE_STORE_TYPES, 'Invalid cache store type'

    return _STORES[store_type].get_cached(identifier)


def uncache(identifier):
    """Uncaches a node.

    :param str identifier: A vocabulary node identifier.

    """
    for store in _STORES.values():
        store.uncache(identifier)



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._codecs.dict.encoder.py
   :copyright: @2015 IPSL (http://ipsl.fr)
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Encodes a term to a python dictionary.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyessv._model import Authority
from pyessv._model import Collection
from pyessv._model import Node
from pyessv._model import Scope
from pyessv._model import Term
from pyessv._utils.compat import str



def encode(instance):
    """Encodes an instance of a domain model class as a dictionary.

    :param pyessv.Node instance: A domain model class instance to be encoded as a dictionary.

    :returns: Instance encoded as a simple dictionary.
    :rtype: dict

    :raises TypeError: If instance is a domain model class instance.

    """
    assert isinstance(instance, Node), 'Invalid type'

    encoders = {
        Authority: _encode_authority,
        Collection: _encode_collection,
        Scope: _encode_scope,
        Term: _encode_term
        }
    encoder = encoders[type(instance)]

    return encoder(instance)


def _encode_authority(instance):
    """Encodes a term authority as a dictionary.

    """
    obj = _encode_node(instance)
    obj['scopes'] = [encode(i) for i in instance]

    return obj


def _encode_scope(instance):
    """Encodes a term scope as a dictionary.

    """
    obj = _encode_node(instance)
    obj['collections'] = [encode(i) for i in instance]

    return obj


def _encode_collection(instance):
    """Encodes a collection as a dictionary.

    """
    obj = _encode_node(instance)
    obj['terms'] = ['{}:{}'.format(i.canonical_name, i.label) for i in instance]
    obj['term_regex'] = instance.term_regex

    return obj


def _encode_term(instance):
    """Encodes a term as a dictionary.

    """
    obj = _encode_node(instance)
    obj['status'] = instance.status
    if bool(instance.parent):
        obj['parent'] = instance.parent.namespace
    if bool(instance.associations):
        obj['associations'] = [i.namespace for i in instance.associations]

    return obj


def _encode_node(instance):
    """Encodes a node instance to a dictionary representation.

    """
    obj = dict()
    obj['_type'] = instance.typekey
    obj['canonical_name'] = instance.canonical_name
    obj['create_date'] = instance.create_date
    obj['namespace'] = instance.namespace
    if bool(instance.label) and instance.label != instance.canonical_name:
        obj['label'] = instance.label
    if bool(instance.raw_name) and instance.raw_name != instance.canonical_name:
        obj['raw_name'] = instance.raw_name
    if bool(instance.data):
        obj['data'] = instance.data
    if bool(instance.description):
        obj['description'] = instance.description
    if bool(instance.alternative_names):
        obj['alternative_names'] = instance.alternative_names
    if bool(instance.url):
        obj['url'] = instance.url

    return obj



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._constants.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Package constants.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import os
import re


# In memory cache type.
CACHE_STORE_MEMORY = 'memory'

# Set of allowed cache types.
CACHE_STORE_TYPES = (
    CACHE_STORE_MEMORY,
    )

# Directory containing vocabulary archive.
DIR_ARCHIVE = os.getenv('PYESSV_ARCHIVE_HOME', os.path.expanduser('~/.esdoc/pyessv-archive'))

# Dictionary encoding.
ENCODING_DICT = 'dict'

# JSON encoding.
ENCODING_JSON = 'json'

# Set of supported encodings.
ENCODING_SET = {
    ENCODING_DICT,
    ENCODING_JSON
}

# Node type key - an authority governing vocabularies.
NODE_TYPEKEY_AUTHORITY = 'authority'

# Node type key - a scope constraining collection of vocabularies.
NODE_TYPEKEY_SCOPE = 'scope'

# Node type key - a collection constraining collection of terms by composing keys from other collections.
NODE_TYPEKEY_COLLECTION = 'collection'

# Node type key - a term.
NODE_TYPEKEY_TERM = 'term'

# Set of allowed name types.
NODE_TYPEKEY_SET = (
  NODE_TYPEKEY_AUTHORITY,
  NODE_TYPEKEY_COLLECTION,
  NODE_TYPEKEY_SCOPE,
  NODE_TYPEKEY_TERM
  )

# Governance state - the term is accepted.
GOVERNANCE_STATUS_ACCEPTED = 'accepted'

# Governance state - the term is obsolete.
GOVERNANCE_STATUS_DEPRECATED = 'obsolete'

# Governance state - the term is pending review.
GOVERNANCE_STATUS_PENDING = 'pending'

# Governance state - the term is rejected.
GOVERNANCE_STATUS_REJECTED = 'rejected'

# Set of allowed governance states.
GOVERNANCE_STATUS_SET = (
    GOVERNANCE_STATUS_PENDING,
    GOVERNANCE_STATUS_ACCEPTED,
    GOVERNANCE_STATUS_REJECTED,
    GOVERNANCE_STATUS_DEPRECATED
    )

# Parsing strictness: canonical-name;
PARSING_STRICTNESS_0 = 0

# Parsing strictness: raw-name;
PARSING_STRICTNESS_1 = 1

# Parsing strictness: canonical-name + raw-name;
PARSING_STRICTNESS_2 = 2

# Parsing strictness: 2 + alternative_names
PARSING_STRICTNESS_3 = 3

# Parsing strictness: 3 + case-insensitive
PARSING_STRICTNESS_4 = 4

# Set of allowed parsing stricness.
PARSING_STRICTNESS_SET = (
    PARSING_STRICTNESS_0,
    PARSING_STRICTNESS_1,
    PARSING_STRICTNESS_2,
    PARSING_STRICTNESS_3,
    PARSING_STRICTNESS_4
    )

# Regular expression for validating a canonical name.
REGEX_CANONICAL_NAME = r'^[a-z0-9\-]*$'

# Standard node fields.
STANDARD_NODE_FIELDS = (
    'alternative_names',
    'canonical_name',
    'create_date',
    'description',
    'label',
    'raw_name',
    'typekey',
    'url',
    )

# Parsing node fields.
PARSING_NODE_FIELDS = (
    'canonical_name',
    'label',
    'raw_name'
    )



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._model.collection.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: A vocabulary collection, e.g. institute-id.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import re

from pyessv._constants import NODE_TYPEKEY_COLLECTION
from pyessv._constants import PARSING_STRICTNESS_0
from pyessv._constants import PARSING_STRICTNESS_1
from pyessv._constants import PARSING_STRICTNESS_2
from pyessv._constants import PARSING_STRICTNESS_3
from pyessv._constants import PARSING_STRICTNESS_4
from pyessv._constants import PARSING_STRICTNESS_SET
from pyessv._constants import REGEX_CANONICAL_NAME
from pyessv._model.node import IterableNode
from pyessv._model.term import Term
from pyessv._utils.compat import basestring
from pyessv._utils.validation import assert_iterable
from pyessv._utils.validation import assert_namespace
from pyessv._utils.validation import assert_regex
from pyessv._utils.validation import assert_string


class Collection(IterableNode):
    """A vocabulary term collection.

    """
    def __init__(self):
        """Instance constructor.

        """
        self.scope = None
        self.terms = []
        self.term_regex = None
        super(Collection, self).__init__(self.terms, NODE_TYPEKEY_COLLECTION)


    @property
    def ancestors(self):
        """Gets ancestors within archive hierarchy.

        """
        return [self.authority, self.scope]


    @property
    def authority(self):
        """Gets governing authority.

        """
        return self.scope.authority


    @property
    def is_virtual(self):
        """Gets flag indicating whether the collection is a virtual one (i.e. simply constrained by a reg-ex).

        """
        return len(self) == 0


    def get_validators(self):
        """Returns set of validators.

        """
        from pyessv._model.scope import Scope

        def _canonical_name():
            assert_string(self.canonical_name)
            assert_regex(self.canonical_name, REGEX_CANONICAL_NAME)

        def _scope():
            assert isinstance(self.scope, Scope)

        def _term_regex():
            assert isinstance(self.term_regex, (basestring, tuple))
            if isinstance(self.term_regex, basestring):
                assert_string(self.term_regex)
            else:
                assert_iterable(self.term_regex, basestring, tuple)
                assert len(self.term_regex) > 1
                assert self.term_regex[0].count('{}') == (len(self.term_regex) - 1)
                for identifier in self.term_regex[1:]:
                    assert_namespace(identifier, min_length=3, max_length=4)

        def _terms():
            assert_iterable(self.terms, Term)

        return super(Collection, self).get_validators() + (
            _canonical_name,
            _scope,
            _term_regex,
            _terms
            )


    def is_matched(self, name, strictness=PARSING_STRICTNESS_2):
        """Gets flag indicating whether a matching term can be found.

        :param str name: A term name to be validated.
        :param int strictness: Strictness level to apply when applying name matching rules.

        """
        assert isinstance(name, basestring), 'Invalid term name'
        assert strictness in PARSING_STRICTNESS_SET, 'Invalid parsing strictness: {}'.format(strictness)

        # Reg-ex match.
        if self.is_virtual:
            if strictness >= PARSING_STRICTNESS_4:
                name = str(name).strip().lower()
            return re.compile(self.term_regex).match(name) is not None

        # Match by term.
        for term in self:
            # match by: canonical_name
            if strictness == PARSING_STRICTNESS_0:
                if name == term.canonical_name:
                    return term

            # match by: raw_name
            elif strictness == PARSING_STRICTNESS_1:
                if name == term.raw_name:
                    return term

            # match by: canonical_name | raw_name
            elif strictness == PARSING_STRICTNESS_2:
                if name in {term.canonical_name, term.raw_name}:
                    return term

            # match by: alternative_name
            if strictness == PARSING_STRICTNESS_3:
                if name in {term.canonical_name, term.raw_name}.union(set(term.alternative_names)):
                    return term

            # match by: all (case-insensitive)
            if strictness == PARSING_STRICTNESS_4:
                name = str(name).strip().lower()
                if name in [i.lower() for i in term.all_names]:
                    return term

        return False



# -*- coding: utf-8 -*-

"""
.. module:: test_model.py

   :copyright: @2013 Earth System Documentation (https://es-doc.org)
   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Executes pyessv model tests.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import nose

import pyessv as LIB

from pyessv._utils.compat import str
import tests.utils as tu
import tests.utils_model as tum



@nose.with_setup(None, tu.teardown)
def test_iterability():
    """Test iterability of domain model.

    """
    def _test(node, keys):
        """Inner test.

        """
        assert iter(node)
        assert len(node) == len(keys)
        for key in keys:
            assert key in node
            assert node[key] is not None


    for node_factory, keys in (
        (tu.create_authority, [tum.SCOPE_NAME]),
        (tu.create_scope, [tum.COLLECTION_01_NAME, tum.COLLECTION_02_NAME, tum.COLLECTION_03_NAME]),
        (tu.create_collection_01, [tum.TERM_01_NAME]),
        (tu.create_collection_02, [tum.TERM_02_NAME]),
        (tu.create_collection_03, [tum.TERM_03_NAME]),
        ):
        desc = 'iterate --> {}'.format(node_factory.__name__[7:])
        tu.init(_test, desc)
        yield _test, node_factory(), keys



# -*- coding: utf-8 -*-

"""
.. module:: handlers.heartbeat.py
   :license: GPL/CeCIL
   :platform: Unix
   :synopsis: ES-DOC pyessv - heartbeat endpoint.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import datetime as dt

import tornado

import pyessv
from pyessv._ws.utils.http import process_request



class HeartbeatRequestHandler(tornado.web.RequestHandler):
    """Operations heartbeat request handler.

    """
    def get(self):
        """HTTP GET handler.

        """
        def _set_output():
            """Sets response to be returned to client.

            """
            self.output = {
                "message": "ES-DOC PYESSV web service is operational @ {}".format(dt.datetime.utcnow()),
            }

        # Process request.
        process_request(self, _set_output)



# -*- coding: utf-8 -*-
"""
.. module:: utils.convertor.py
   :license: GPL/CeCIL
   :platform: Unix
   :synopsis: Conversion utility functions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import collections
import datetime
import json
import re
import time


# Set of types to be ignored when jsonifying.
_IGNOREABLE = (int, float, long, type(None), unicode)

# Set of unicodeable types used in jsonifying.
_UNICODEABLE = (basestring, datetime.datetime)

# Values considered to be abbreviations.
_ABBREVIATIONS = ("id", "uid", "uuid")

# Default separator.
_DEFAULT_SEPARATOR = "_"

# Set of db coumns that are ignored when converting db entity instances.
_IGNORED_DB_COLUMNS = {'id', 'row_create_date', 'row_update_date'}


def to_dict(data, key_convertor=None):
    """Converts input data to a dictionary.

    :param object data: Data to be converted.
    :param func key_convertor: Dictionary key convertor.

    :returns: Converted data.
    :rtype: object

    """
    # Ignoreable types.
    if isinstance(data, _IGNOREABLE):
        return data

    # Unicodeable types.
    elif isinstance(data, _UNICODEABLE):
        return unicode(data)

    # Dictionaries.
    elif isinstance(data, collections.Mapping):
        return {k if key_convertor is None else key_convertor(k):
                to_dict(v, key_convertor) for k, v in data.iteritems()}

    # Collections.
    elif isinstance(data, collections.Iterable):
        return [to_dict(i, key_convertor) for i in data]

    else:
        return data


def to_namedtuple(obj, key_convertor=None):
    """Converts a dictionary to a named tuple.

    :param dict obj: Dictionary for conversion.
    :param func key_convertor: Dictionary key convertor.

    :returns: A named tuple.
    :rtype: namedtuple

    """
    obj = to_dict(obj, key_convertor)
    # Convert sub-dictionaries.
    for k, v in obj.items():
        if isinstance(v, dict):
            obj[k] = to_namedtuple(v, key_convertor)

    return collections.namedtuple('_Class', obj.keys())(**obj)


def to_json(data):
    """Converts input dictionary to json.

    :param dict data: Data in dictionary format.
    :param bool sort_keys: Flag indicating whether the dictionary keys will be sorted.

    :returns: JSON encoded string.
    :rtype: str

    """
    return json.dumps(to_dict(data, key_convertor=to_camel_case))


def json_file_to_dict(fpath, key_convertor=None):
    """Converts a json file to a dictionary.

    :param str fpath: A json file path.
    :param func key_convertor: Dictionary key convertor.

    :returns: A dictionary.
    :rtype: dict

    """
    with open(fpath, 'r') as fstream:
        fdata = json.loads(fstream.read())

    return to_dict(fdata, key_convertor)


def json_file_to_namedtuple(fpath, key_convertor=None):
    """Converts a json file to a namedtuple.

    :param str fpath: A json file path.
    :param func key_convertor: Dictionary key convertor.

    :returns: A namedtuple.
    :rtype: namedtuple

    """
    return to_namedtuple(json_file_to_dict(fpath, key_convertor))


def to_pascal_case(target, separator=_DEFAULT_SEPARATOR):
    """Converts a string to pascal case.

    :param str target: A string to be converted.
    :param str separator: A separator used to split target string into parts.

    :returns: The target string converted to pascal case.
    :rtype: str

    """
    result = ''
    if target is not None and len(target):
        if target[0:len(separator)] == separator:
            result = separator
        for text in target.split(separator):
            if text.lower() in _ABBREVIATIONS:
                result += text.upper()
            elif (len(text) > 0):
                result += text[0].upper()
                if (len(text) > 1):
                    result += text[1:]
    return result


def to_camel_case(target, separator=_DEFAULT_SEPARATOR):
    """Converts a string to camel case.

    :param str target: A string to be converted.
    :param str separator: A separator used to split target string into parts.

    :returns: The target string converted to camel case.
    :rtype: str

    """
    result = ''
    if target is not None and len(target):
        text = to_pascal_case(target, separator)
        # Preserve initial separator
        if text[0:len(separator)] == separator:
            result += separator
            text = text[len(separator):]

        # Lower case abbreviations.
        if text.lower() in _ABBREVIATIONS:
            result += text.lower()

        # Lower case initial character.
        elif len(text):
            result += text[0].lower()
            result += text[1:]
    return result


def to_spaced_case(target, separator=_DEFAULT_SEPARATOR):
    """Converts a string to spaced case.

    :param str target: A string for conversion.

    :returns: A string converted to spaced case.
    :rtype: str

    """
    if target is None:
        return ""
    elif separator is not None and len(target.split(separator)) > 1:
        return " ".join(target.split(separator))
    elif target.find(" ") == -1:
        return re.sub("([A-Z])", r" \g<0>", target).strip()
    else:
        return target


def to_underscore_case(target):
    """Converts a string to underscore case.

    :param str target: A string for conversion.

    :returns: A string converted to underscore case, e.g. account_number.
    :rtype: str

    """
    if target is None or not len(target):
        return unicode()

    result = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', target)
    result = re.sub('([a-z0-9])([A-Z])', r'\1_\2', result)

    return result.lower()



def now_to_timestamp(offset=0):
    """Returns a timestamp from datatime.datetime.now().

    """
    now = time.time()
    localtime = time.localtime(now)
    milliseconds = '%03d' % int((now - int(now)) * 1000)
    ts = time.strftime('%Y%m%d%H%M%S', localtime) + milliseconds

    return int(ts) + offset



# -*- coding: utf-8 -*-
"""
.. module:: utils.http_security.py
   :license: GPL/CeCIL
   :platform: Unix
   :synopsis: Miscellaneous utility functions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyessv._ws.utils import exceptions



def secure_request(handler):
    """Enforces request level security policy (if necessary).

    :param utils.http.HTTPRequestHandler handler: An HTTP request handler.

    """
    # Nothing to do at this point in time.
    pass



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._cache.memory.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Simple in-memory cache store.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyessv._model import NODE_TYPES
from pyessv._model import Authority
from pyessv._utils.compat import str



# Cached loaded vocabulary authorities objects.
_DATA = {}


def cache(node):
    """Caches a vocabulary node.

    :param pyeesv.Node: Node to be cached.

    """
    _DATA[node.namespace] = node
    try:
        iter(node)
    except TypeError:
        return
    else:
        for subnode in node:
            cache(subnode)


def get_cached(cache_filter):
    """Returns a cached node.

    :param str|class filter: Cache filter expression.

    :returns: A pointer to a vocabulary node or a set of nodes.
    :rtype: pyessv.Node | list[Authority]

    """
    if cache_filter in _DATA:
        return _DATA[cache_filter]
    elif cache_filter in NODE_TYPES:
        return [i for i in _DATA.values() if isinstance(i, cache_filter)]
    elif cache_filter is None:
        return get_cached(Authority)


def uncache(identifier):
    """Uncaches a node.

    :param str identifier: A vocabulary node identifier.

    """
    try:
        del _DATA[identifier]
    except KeyError:
        pass



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._model._parser_template.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: A vocabulary constrained template parser, e.g. a dataset identifier.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyessv._exceptions import TemplateParsingError
from pyessv._model.collection import Collection
from pyessv._model.term import Term
from pyessv._utils.compat import  basestring



class TemplateParser(object):
    """A vocabulary template parser.

    """
    def __init__(self, template, collections, strictness, seperator='.'):
        """Instance constructor.

        :param str template: Identifier template.
        :param list collections: pyessv collection identifiers.
        :param int strictness: Strictness level to apply when applying name matching rules.
        :param str seprarator: Seperator to apply when parsing.

        """
        from pyessv._loader import load

        self.seperator = seperator
        self.template_parts = template.split(seperator)
        self.template = template
        self.strictness = strictness

        # Inject pyessv collections into template.
        collection_idx = 0
        for idx, part in enumerate(self.template_parts):
            if part == '{}':
                self.template_parts[idx] = load(collections[collection_idx])
                collection_idx += 1


    def parse(self, val):
        """Parses a val against a template.

        :returns: Set of terms extracted during parsing process.

        """
        # Verify that number of parts is equal.
        parts = val.split(self.seperator)
        if len(parts) != len(self.template_parts):
            raise TemplateParsingError('Number of elements is invalid: {}: is {}, expected {}'.format(val, len(parts), len(self.template_parts)))

        # Iterate template.
        terms = set()
        for template_part, name in [(self.template_parts[i], j) for i, j in enumerate(parts)]:
            # Verify constant match.
            if isinstance(template_part, basestring):
                if template_part != name:
                    raise TemplateParsingError('{} :: {}'.format(name, val))
                continue

            # Verify collection match.
            collection = template_part
            term = collection.is_matched(name, self.strictness)
            if term == False:
                raise TemplateParsingError('vocab={} :: strictness={} :: value={}'.format(collection, self.strictness, val))
            if isinstance(term, Term):
                terms.add(term)

        return terms



# -*- coding: utf-8 -*-
"""

.. module:: schemas.loader.py
   :license: GPL/CeCIL
   :platform: Unix
   :synopsis: ES-DOC Errata - endpoint validation schema cache loader.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import json
import os

from pyessv._ws.utils import constants
from pyessv._ws.schemas import extender



def load(typeof, endpoint):
	"""Returns a loaded schema.

	"""
	fpath = _get_fpath(typeof, endpoint)
	if os.path.exists(fpath):
		try:
			with open(fpath, 'r') as fstream:
				schema = json.loads(fstream.read())
		except Exception as err:
			print endpoint, typeof, err
			pass
		else:
			extender.extend(schema, typeof, endpoint)
			return schema


def _get_fpath(typeof, endpoint):
	"""Returns schema file path.

	"""
	endpoint = constants.DEFAULT_ENDPOINT if endpoint == '/' else endpoint
	fname = "{}.json".format(endpoint[1:].replace("/", "."))
	fpath = os.path.join(os.path.dirname(__file__), typeof)

	return os.path.join(fpath, fname)



# -*- coding: utf-8 -*-

"""
.. module:: pyessv.utils.validation.py
   :copyright: Copyright "December 01, 2016', IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Validation utility functions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import inspect
import re

from pyessv._constants import PARSING_NODE_FIELDS
from pyessv._utils.compat import basestring
from pyessv._utils.compat import str
from pyessv._utils.compat import urlparse


def assert_iterable(val, modifier, typeof=list):
    """Asserts an iterable value.

    """
    assert isinstance(val, typeof)
    for i in val:
        if inspect.isfunction(modifier):
            modifier(i)
        else:
            assert isinstance(i, modifier)


def assert_regex(val, reg_ex):
    """Asserts a string value.

    """
    assert re.compile(reg_ex).match(val) is not None


def assert_string(val):
    """Asserts a string value.

    """
    assert isinstance(val, basestring)
    assert len(val.strip()) > 0


def assert_url(val):
    """Asserts a url value.

    """
    assert_string(val)
    url = urlparse(val)
    assert url.netloc and url.scheme


def assert_namespace(identifier, min_length=1, max_length=4):
    """Asserts a namespace.

    """
    assert_string(identifier)
    parts = str(identifier).split(':')
    assert len(parts) >= min_length and len(parts) <= max_length
    for part in parts:
        assert_string(part)


def assert_pattern(val, info):
    from pyessv._loader import load

    pattern = info[0]
    collections = []
    for defn in info[1:]:
        parts = defn.split(':')
        field = parts[-1] if parts[-1] in PARSING_NODE_FIELDS else 'canonical_name'
        collection = load(':'.join(parts[0:3]))
        collections.append((collection, field))



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._parsers.cmip6_dataset_id.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Encapsulates parsing of a CMIP5 dataset identifier.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>

"""
from pyessv._factory import create_template_parser
from pyessv._constants import PARSING_STRICTNESS_1



# Template extracted from esgf ini file (for reference purpose only).
_INI_PATTERN = 'cmip5.%(product)s.%(institute)s.%(model)s.%(experiment)s.%(time_frequency)s.%(realm)s.%(cmor_table)s.%(ensemble)s'

# Test identifier (for reference purpose only).
_TEST_IDENTIFIER = 'cmip5.output2.IPSL.IPSL-CM5A-LR.historicalMisc.mon.ocean.Omon.r2i1p1.v20150504'

# Template that identifiers must conform to.
_TEMPLATE = 'cmip5.{}.{}.{}.{}.{}.{}.{}.{}'

# Collections injected into template.
_COLLECTIONS = (
    'wcrp:cmip5:product',
    'wcrp:cmip5:institute',
    'wcrp:cmip5:model',
    'wcrp:cmip5:experiment',
    'wcrp:cmip5:time-frequency',
    'wcrp:cmip5:realm',
    'wcrp:cmip5:cmor-table',
    'wcrp:cmip5:ensemble'
    )

# Instantiated & cached parser instance.
_PARSER = None


def parse(identifier):
    """Parses a CMIP6 dataset identifier.

    """
    # Instantiate parser JIT.
    global _PARSER
    if _PARSER is None:
        _PARSER = create_template_parser(_TEMPLATE, _COLLECTIONS, PARSING_STRICTNESS_1)

    # Strip version suffix.
    if '#' in identifier:
      identifier = identifier.split('#')[0]

    return _PARSER.parse(identifier.split('#')[0])



# -*- coding: utf-8 -*-
"""
.. module:: pyessv.__init__.py

   :copyright: @2015 IPSL (http://ipsl.fr)
   :license: GPL / CeCILL
   :platform: Unix
   :synopsis: Python Earth Science Standard Vocabulary library intializer.

.. moduleauthor:: IPSL (ES-DOC) <dev@esdocumentation.org>

"""
__title__ = 'pyessv'
__version__ = '0.6.6.0'
__author__ = 'ES-DOC'
__license__ = 'GPL'
__copyright__ = 'Copyright 2017 ES-DOC'


from pyessv._archive import archive

from pyessv._cache import get_cached

from pyessv._codecs import decode
from pyessv._codecs import encode

from pyessv._constants import DIR_ARCHIVE
from pyessv._constants import ENCODING_DICT
from pyessv._constants import ENCODING_JSON
from pyessv._constants import NODE_TYPEKEY_AUTHORITY
from pyessv._constants import NODE_TYPEKEY_COLLECTION
from pyessv._constants import NODE_TYPEKEY_SCOPE
from pyessv._constants import NODE_TYPEKEY_TERM
from pyessv._constants import GOVERNANCE_STATUS_ACCEPTED
from pyessv._constants import GOVERNANCE_STATUS_DEPRECATED
from pyessv._constants import GOVERNANCE_STATUS_PENDING
from pyessv._constants import GOVERNANCE_STATUS_REJECTED
from pyessv._constants import PARSING_STRICTNESS_0
from pyessv._constants import PARSING_STRICTNESS_1
from pyessv._constants import PARSING_STRICTNESS_2
from pyessv._constants import PARSING_STRICTNESS_3
from pyessv._constants import PARSING_STRICTNESS_4
from pyessv._constants import REGEX_CANONICAL_NAME

from pyessv._exceptions import InvalidAssociationError
from pyessv._exceptions import TemplateParsingError
from pyessv._exceptions import ParsingError
from pyessv._exceptions import ValidationError

from pyessv._factory import create_authority
from pyessv._factory import create_collection
from pyessv._factory import create_template_parser
from pyessv._factory import create_scope
from pyessv._factory import create_term

from pyessv._factory_for_testing import get_test_datasets

from pyessv._governance import accept
from pyessv._governance import deprecate
from pyessv._governance import destroy
from pyessv._governance import reject
from pyessv._governance import reset

from pyessv._initializer import init

from pyessv._loader import load_random
from pyessv._loader import load

from pyessv._model import Authority
from pyessv._model import Collection
from pyessv._model import Scope
from pyessv._model import Term

from pyessv._parser import parse
from pyessv._parsers import parse_dataset_identifer
from pyessv._parsers import parse_dataset_identifers

from pyessv._utils.logger import log
from pyessv._utils.logger import log_error
from pyessv._utils.logger import log_warning

from pyessv._validation import get_errors
from pyessv._validation import is_valid
from pyessv._validation import validate

from pyessv._ws import app as web_service


# Auto-initialize.
init()



# -*- coding: utf-8 -*-

"""
.. module:: utils_cv.py

   :copyright: @2013 Earth System Documentation (https://es-doc.org)
   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Exposes other test utility functions.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import inspect
import os
import shutil

import pyessv as LIB
from pyessv._cache import uncache
from tests.utils_assert import assert_objects


# Test authority.
AUTHORITY = None
AUTHORITY_DESCRIPTION = 'authority-description'
AUTHORITY_NAME = 'authority'
AUTHORITY_NAMESPACE = AUTHORITY_NAME
AUTHORITY_ALTERNATIVE_NAMES = ['authority-alternative-name-1', 'authority-alternative-name-2']
AUTHORITY_URL = 'https://github.com/ES-DOC/pyesssv-archive/{}'.format(AUTHORITY_NAME)


def create_authority():
    """Creates & returns a test authority.

    """
    global AUTHORITY

    if AUTHORITY is not None:
        return AUTHORITY

    AUTHORITY = LIB.create_authority(
        name=AUTHORITY_NAME,
        description=AUTHORITY_DESCRIPTION,
        alternative_names=AUTHORITY_ALTERNATIVE_NAMES,
        url=AUTHORITY_URL
        )
    create_scope()

    return AUTHORITY


# Test scope.
SCOPE = None
SCOPE_DESCRIPTION = 'scope-description'
SCOPE_NAME = 'scope'
SCOPE_NAMESPACE = ":".join([AUTHORITY_NAME, SCOPE_NAME])
SCOPE_ALTERNATIVE_NAMES = ['scope-alternative-name-1', 'scope-alternative-name-2']
SCOPE_URL = '{}/{}'.format(AUTHORITY_URL, SCOPE_NAME)


def create_scope():
    """Creates & returns a test scope.

    """
    global SCOPE

    if SCOPE is not None:
        return SCOPE

    SCOPE = LIB.create_scope(
        authority=AUTHORITY or create_authority(),
        name=SCOPE_NAME,
        description=SCOPE_DESCRIPTION,
        alternative_names=SCOPE_ALTERNATIVE_NAMES,
        url=SCOPE_URL
        )
    create_collection_01()
    create_collection_02()
    create_collection_03()

    return SCOPE


# Test collection - 01.
COLLECTION_01 = None
COLLECTION_01_DESCRIPTION = 'collection-01-description'
COLLECTION_01_NAME = 'collection-01'
COLLECTION_01_NAMESPACE = ":".join([AUTHORITY_NAME, SCOPE_NAME, COLLECTION_01_NAME])
COLLECTION_01_ALTERNATIVE_NAMES = ['collection-01-alternative-name-1', 'collection-01-alternative-name-2']
COLLECTION_01_URL = '{}/{}'.format(SCOPE_URL, COLLECTION_01_NAME)


def create_collection_01():
    """Creates & returns a test collection.

    """
    global COLLECTION_01

    if COLLECTION_01 is not None:
        return COLLECTION_01

    COLLECTION_01 = LIB.create_collection(
        scope=SCOPE or create_scope(),
        name=COLLECTION_01_NAME,
        description=COLLECTION_01_DESCRIPTION,
        alternative_names=COLLECTION_01_ALTERNATIVE_NAMES,
        url=COLLECTION_01_URL
        )
    create_term_01()

    return COLLECTION_01


# Test collection - 02.
COLLECTION_02 = None
COLLECTION_02_DESCRIPTION = 'collection-02-description'
COLLECTION_02_NAME = 'collection-02'
COLLECTION_02_NAMESPACE = ":".join([AUTHORITY_NAME, SCOPE_NAME, COLLECTION_02_NAME])
COLLECTION_02_ALTERNATIVE_NAMES = ['collection-02-alternative-name-1', 'collection-02-alternative-name-2']
COLLECTION_02_TERM_REGEX = r'^[A-z0-9\-]*$'
COLLECTION_02_URL = '{}/{}'.format(SCOPE_URL, COLLECTION_02_NAME)


def create_collection_02():
    """Creates & returns a test collection.

    """
    global COLLECTION_02

    if COLLECTION_02 is not None:
        return COLLECTION_02

    COLLECTION_02 = LIB.create_collection(
        scope=SCOPE or create_scope(),
        name=COLLECTION_02_NAME,
        description=COLLECTION_02_DESCRIPTION,
        alternative_names=COLLECTION_02_ALTERNATIVE_NAMES,
        term_regex=COLLECTION_02_TERM_REGEX,
        url=COLLECTION_02_URL
        )
    create_term_02()

    return COLLECTION_02


# Test composite collection.
COLLECTION_03 = None
COLLECTION_03_DESCRIPTION = 'collection-03-description'
COLLECTION_03_NAME = 'collection-03'
COLLECTION_03_NAMESPACE = ":".join([AUTHORITY_NAME, SCOPE_NAME, COLLECTION_03_NAME])
COLLECTION_03_ALTERNATIVE_NAMES = ['collection-03-alternative-name-1', 'collection-03-alternative-name-2']
COLLECTION_03_TERM_REGEX = r'^[A-z0-9\-]*$'
COLLECTION_03_URL = '{}/{}'.format(SCOPE_URL, COLLECTION_03_NAME)


def create_collection_03():
    """Creates & returns a test collection.

    """
    global COLLECTION_03

    if COLLECTION_03 is not None:
        return COLLECTION_03

    COLLECTION_03 = LIB.create_collection(
        scope=SCOPE or create_scope(),
        name=COLLECTION_03_NAME,
        description=COLLECTION_03_DESCRIPTION,
        alternative_names=COLLECTION_03_ALTERNATIVE_NAMES,
        term_regex=COLLECTION_03_TERM_REGEX,
        url=COLLECTION_03_URL
        )
    create_term_03()

    return COLLECTION_03


# Test term - 01.
TERM_01 = None
TERM_01_DESCRIPTION = 'term-01-description'
TERM_01_NAME = 'term-01'
TERM_01_NAMESPACE = ":".join([AUTHORITY_NAME, SCOPE_NAME, COLLECTION_01_NAME, TERM_01_NAME])
TERM_01_ALTERNATIVE_NAMES = ['term-01-alternative-name-1', 'term-01-alternative-name-2']
TERM_01_URL = '{}/{}'.format(COLLECTION_01_URL, TERM_01_NAME)


def create_term_01(collection=None):
    """Creates & returns a test term.

    """
    global TERM_01

    if TERM_01 is not None:
        return TERM_01

    TERM_01 = LIB.create_term(
        collection=collection or COLLECTION_01 or create_collection_01(),
        name=TERM_01_NAME,
        description=TERM_01_DESCRIPTION,
        alternative_names=TERM_01_ALTERNATIVE_NAMES,
        url=TERM_01_URL
        )

    return TERM_01


# Test term - 02.
TERM_02 = None
TERM_02_DESCRIPTION = 'term-02-description'
TERM_02_NAME = 'term-02'
TERM_02_NAMESPACE = ":".join([AUTHORITY_NAME, SCOPE_NAME, COLLECTION_02_NAME, TERM_02_NAME])
TERM_02_ALTERNATIVE_NAMES = ['term-02-alternative-name-1', 'term-02-alternative-name-2']
TERM_02_URL = '{}/{}'.format(COLLECTION_02_URL, TERM_02_NAME)


def create_term_02(collection=None):
    """Creates & returns a test term.

    """
    global TERM_02

    if TERM_02 is not None:
        return TERM_02

    TERM_02 = LIB.create_term(
        collection=collection or COLLECTION_02 or create_collection_02(),
        name=TERM_02_NAME,
        description=TERM_02_DESCRIPTION,
        alternative_names=TERM_02_ALTERNATIVE_NAMES,
        url=TERM_02_URL
        )

    return TERM_02


# Test term - 03.
TERM_03 = None
TERM_03_DESCRIPTION = 'term-03-description'
TERM_03_NAME = '{}-TESTING-{}'.format(TERM_01_NAME, TERM_02_NAME)
TERM_03_NAMESPACE = ":".join([AUTHORITY_NAME, SCOPE_NAME, COLLECTION_03_NAME, TERM_03_NAME])
TERM_03_ALTERNATIVE_NAMES = ['term-03-alternative-name-1', 'term-03-alternative-name-2']
TERM_03_URL = '{}/{}'.format(COLLECTION_03_URL, TERM_03_NAME)


def create_term_03(collection=None):
    """Creates & returns a test term.

    """
    global TERM_03

    if TERM_03 is not None:
        return TERM_03

    TERM_03 = LIB.create_term(
        collection=collection or COLLECTION_03 or create_collection_03(),
        name=TERM_03_NAME,
        description=TERM_03_DESCRIPTION,
        alternative_names=TERM_03_ALTERNATIVE_NAMES,
        url=TERM_03_URL
        )

    return TERM_03


def create_test_entities():
    """Returns tuple of test entities.

    """
    return \
        create_term_01(), \
        create_term_02(), \
        create_term_03(), \
        create_collection_01(), \
        create_collection_02(), \
        create_collection_03(), \
        create_scope(), \
        create_authority()


def init(func, desc=None):
    """Initializes a test function prior to be executed.

    """
    if desc is None:
        desc = inspect.getdoc(func)
    desc = desc.strip()
    if desc[-1] == '.':
        desc = desc[:-1]
    desc = desc[0].lower() + desc[1:]
    func.description = 'pyessv-tests: {}'.format(desc)


def setup():
    """Initialises test objects.

    """
    teardown()
    for func in (
        create_authority,
        create_scope,
        create_collection_01,
        create_collection_02,
        create_collection_03,
        create_term_01,
        create_term_02,
        create_term_03
        ):
        func()


def teardown():
    """Performs teardown functions after running a test.

    """
    global AUTHORITY
    global SCOPE
    global COLLECTION_01
    global COLLECTION_02
    global COLLECTION_03
    global TERM_01
    global TERM_02
    global TERM_03

    uncache(AUTHORITY)

    AUTHORITY = None
    SCOPE = None
    COLLECTION_01 = None
    COLLECTION_02 = None
    COLLECTION_03 = None
    TERM_01 = None
    TERM_02 = None
    TERM_03 = None

    try:
        shutil.rmtree(os.path.join(LIB.DIR_ARCHIVE, AUTHORITY_NAME))
    except OSError:
        pass



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._model.scope.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: A vocabulary scope, e.g. CMIP6.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyessv._constants import NODE_TYPEKEY_SCOPE
from pyessv._constants import REGEX_CANONICAL_NAME
from pyessv._model.collection import Collection
from pyessv._model.node import IterableNode
from pyessv._utils.validation import assert_iterable
from pyessv._utils.validation import assert_regex
from pyessv._utils.validation import assert_string



class Scope(IterableNode):
    """A scope managed by an authority.

    """
    def __init__(self):
        """Instance constructor.

        """
        self.authority = None
        self.collections = []
        super(Scope, self).__init__(self.collections, NODE_TYPEKEY_SCOPE)


    @property
    def ancestors(self):
        """Gets ancestors within archive hierarchy.

        """
        return [self.authority]


    def get_validators(self):
        """Returns set of validators.

        """
        from pyessv._model.authority import Authority

        def _authority():
            assert isinstance(self.authority, Authority)

        def _canonical_name():
            assert_string(self.canonical_name)
            assert_regex(self.canonical_name, REGEX_CANONICAL_NAME)

        def _collections():
            assert_iterable(self.collections, Collection)

        return super(Scope, self).get_validators() + (
            _authority,
            _canonical_name,
            _collections
            )



# -*- coding: utf-8 -*-

"""
.. module:: dict.decoder.py
   :copyright: @2015 IPSL (http://ipsl.fr)
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Decodes a term from a python dictionary.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import arrow

from pyessv._constants import NODE_TYPEKEY_AUTHORITY
from pyessv._constants import NODE_TYPEKEY_COLLECTION
from pyessv._constants import NODE_TYPEKEY_SCOPE
from pyessv._constants import NODE_TYPEKEY_SET
from pyessv._constants import NODE_TYPEKEY_TERM
from pyessv._model import Authority
from pyessv._model import Collection
from pyessv._model import Scope
from pyessv._model import Term
from pyessv._utils import convert
from pyessv._utils.compat import str



def decode(obj):
    """Decodes a term from a dictionary.

    :param dict obj: Dictionary to be decoded.

    :returns: Decoded term.
    :rtype: pyessv.Term

    """
    assert '_type' in obj, 'Invalid type key'
    assert obj['_type'] in NODE_TYPEKEY_SET, 'Invalid type key'

    decoders = {
        NODE_TYPEKEY_AUTHORITY: _decode_authority,
        NODE_TYPEKEY_COLLECTION: _decode_collection,
        NODE_TYPEKEY_SCOPE: _decode_scope,
        NODE_TYPEKEY_TERM: _decode_term
        }
    decoder = decoders[obj['_type']]

    return decoder(obj)


def _decode_authority(obj):
    """Decodes a termset from a dictionary.

    """
    instance = _decode_node(obj, Authority)
    for scope in [decode(i) for i in obj['scopes']]:
        scope.authority = instance
        instance.scopes.append(scope)

    return instance


def _decode_scope(obj):
    """Decodes a termset from a dictionary.

    """
    instance = _decode_node(obj, Scope)
    for collection in [decode(i) for i in obj['collections']]:
        collection.scope = instance
        instance.collections.append(collection)

    return instance


def _decode_collection(obj):
    """Decodes a collection from a dictionary.

    """
    instance = _decode_node(obj, Collection)
    instance.term_regex = obj.get('term_regex')

    return instance


def _decode_term(obj):
    """Decodes a term from a dictionary.

    """
    instance = _decode_node(obj, Term)
    instance.associations = obj.get('associations', [])
    instance.status = obj['status']

    return instance


def _decode_node(obj, typeof):
    """Decodes a node instance from a dictionary representation.

    """
    instance = typeof()
    instance.alternative_names = obj.get('alternative_names', [])
    instance.create_date = arrow.get(obj['create_date']).datetime
    instance.data = obj.get('data', dict())
    instance.description = obj.get('description')
    instance.label = obj.get('label', obj['canonical_name'])
    instance.canonical_name = obj['canonical_name']
    instance.raw_name = obj.get('raw_name', obj['canonical_name'])
    instance.url = obj.get('url')

    return instance



# -*- coding: utf-8 -*-

"""
.. module:: handlers.retrieve.py
   :license: GPL/CeCIL
   :platform: Unix
   :synopsis: ES-DOC pyessv - retrieve node endpoint.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from collections import OrderedDict

import tornado

import pyessv
from pyessv._ws.utils import constants
from pyessv._ws.utils.http import process_request




# Query parameters.
_PARAM_INCLUDE_META = 'includeMeta'


class RetrieveRequestHandler(tornado.web.RequestHandler):
    """Retrieve node request handler.

    """
    def set_default_headers(self):
        """Set HTTP headers at the beginning of the request.

        """
        self.set_header(constants.HTTP_HEADER_Access_Control_Allow_Origin, "*")


    def get(self):
        """HTTP GET handler.

        """
        def _set_output():
            """Sets response to be returned to client.

            """
            # Set include meta section flag.
            include_meta = self.get_argument(_PARAM_INCLUDE_META, 'false') == 'true'

            # Sets vocabulary identifier.
            identifier = ':'.join([i.strip().lower().replace('_', '-') for i in self.request.path.split('/')[3:]])

            # Set output to be returned to client.
            if len(identifier) == 0:
                self.output = {
                    'data': [_encode(i, include_meta) for i in pyessv.load()]
                }
            else:
                self.output = {
                    'data': _encode(pyessv.load(identifier), include_meta)
                }


        # Process request.
        process_request(self, [
            _set_output
            ])


def _encode(node, include_meta):
    """Encodes output according to node type.

    """
    obj = _encode_node(node, include_meta)
    if isinstance(node, pyessv.Authority):
        obj['scopes'] = [_encode(i, include_meta) for i in node.scopes]
    elif isinstance(node, pyessv.Scope):
        obj['collections'] = [_encode(i, include_meta) for i in node.collections]
    elif isinstance(node, pyessv.Collection):
        obj['terms'] = [_encode(i, include_meta) for i in node.terms]
        obj['term_regex'] = node.term_regex
    elif isinstance(node, pyessv.Term):
        obj['data'] = node.data
        obj['status'] = node.status

    return obj


def _encode_node(node, include_meta):
    """Returns encoded node base class.

    """
    obj = {
        'alternative_names': node.alternative_names,
        'canonical_name': node.canonical_name,
        'description': node.description,
        'label': node.label,
        'raw_name': node.raw_name,
        'url': node.url
    }
    if include_meta:
        obj['meta'] = {
            'create_date': node.create_date,
            'namespace': node.namespace,
            'typekey': node.typekey
        }

    return obj



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._model.node.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: A node within the pyessv domain model.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import datetime

import arrow

from pyessv._constants import NODE_TYPEKEY_SET
from pyessv._utils.compat import str
from pyessv._utils.formatter import format_io_name
from pyessv._utils.validation import assert_iterable
from pyessv._utils.validation import assert_string
from pyessv._utils.validation import assert_url


class Node(object):
    """A node within the pyessv domain model.

    """
    def __init__(self, typekey):
        """Instance constructor.

        """
        self.canonical_name = None
        self.create_date = None
        self.data = None
        self.description = None
        self.label = None
        self.raw_name = None
        self.alternative_names = list()
        self.typekey = typekey
        self.url = None


    def __repr__(self):
        """Instance representation.

        """
        return self.namespace


    def __getattr__(self, name):
        """Instance attribute getter.

        """
        try:
            return self.data[name]
        except KeyError:
            try:
                return self.data[name.replace('_', '-')]
            except KeyError:
                raise AttributeError('{} unknown attribute'.format(name))


    @property
    def name(self):
        """Helper attribute to return canonical_name.

        """
        return self.canonical_name


    @property
    def all_names(self):
        """Returns all term names.

        """
        result = [self.canonical_name, self.raw_name] + self.alternative_names
        result = [t for t in result if t is not None and len(t) > 0]

        return set(sorted(result))


    @property
    def hierarchy(self):
        """Gets hierachy within archive.

        """
        return self.ancestors + [self]


    @property
    def namespace(self):
        """Gets hierachy within archive.

        """
        return ":".join([i.canonical_name for i in self.hierarchy])


    @property
    def io_name(self):
        """Returns name formatted for I/O operations.

        """
        return format_io_name(self.canonical_name)


    def get_validators(self):
        """Returns set of validators.

        """
        def _alternative_names():
            assert_iterable(self.alternative_names, assert_string)

        def _create_date():
            assert isinstance(self.create_date, datetime.datetime)

        def _data():
            if self.data is not None:
                assert isinstance(self.data, dict)

        def _description():
            if isinstance(self, IterableNode):
                assert_string(self.description)
            elif self.description is not None:
                assert_string(self.description)

        def _label():
            if self.label is not None:
                assert_string(self.label)

        def _typekey():
            assert self.typekey in NODE_TYPEKEY_SET

        def _url():
            if self.url is not None:
                assert_url(self.url)

        return (
            _alternative_names,
            _create_date,
            _data,
            _description,
            _label,
            _typekey,
            _url
            )


class IterableNode(Node):
    """An iterable node within the pyessv domain model.

    """
    def __init__(self, items, typekey):
        """Instance constructor.

        """
        self._items = items
        super(IterableNode, self).__init__(typekey)


    def __add__(self, other):
        """Add operator.

        """
        return self._items + other._items


    def __contains__(self, key):
        """Instance membership predicate.

        """
        return self[key] is not None


    def __getattr__(self, name):
        """Instance attribute getter.

        """
        try:
            return self[name.lower().replace('_', '-')]
        except KeyError:
            raise AttributeError('{} unknown attribute'.format(name))


    def __getitem__(self, key):
        """Returns a child section item.

        """
        # Match against a canonical name.
        for item in self:
            if item.canonical_name == key:
                return item

        # Match against a raw name.
        for item in self:
            if key == item.raw_name:
                return item

        # Match against an alternative name.
        for item in self:
            if key in item.alternative_names:
                return item


    def __iter__(self):
        """Instance iterator initializer.

        """
        return iter(sorted(self._items,
                           key=lambda i: i if isinstance(i, basestring) else i.canonical_name))


    def __len__(self):
        """Returns number of items in managed collection.

        """
        return len(self._items)



# -*- coding: utf-8 -*-
"""
.. module:: utils.http.py
   :license: GPL/CeCIL
   :platform: Unix
   :synopsis: HTTP utility functions, particulary process_request.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import pyessv

from pyessv._ws.utils import logger
from pyessv._ws.utils.constants import *
from pyessv._ws.utils.convertor import to_dict
from pyessv._ws.utils.convertor import to_camel_case
from pyessv._ws.utils.exceptions import ERROR_CODES
from pyessv._ws.utils.http_security import secure_request
from pyessv._ws.utils.http_validator import validate_request



def process_request(handler, tasks, error_tasks=None):
    """Invokes a set of HTTP request processing tasks.

    :param HTTPRequestHandler handler: Request processing handler.
    :param list tasks: Collection of processing tasks.
    :param list error_tasks: Collection of error processing tasks.

    """
    # Extend tasksets.
    tasks = _get_tasks([_log_begin, secure_request, validate_request], tasks, [_log_success, _write_success])
    error_tasks = _get_tasks([], error_tasks or [], [_log_error, write_error])

    # Invoke tasksets.
    for task in tasks:
        try:
            _invoke_task(handler, task)
        except Exception as err:
            try:
                for task in error_tasks:
                    _invoke_task(handler, task, err)
            except:
                # suppress error processing exceptions
                pass
            break


def _log(handler, msg, is_error=False):
    """Logs an error response.

    """
    msg = "[{}]: --> {}".format(id(handler), msg)
    if is_error:
        logger.log_web_error(msg)
    else:
        logger.log_web(msg)


def _log_begin(handler):
    """Logs beginning of request processing.

    """
    msg = "[{0}]: executing --> {1}"
    msg = msg.format(id(handler), handler)
    logger.log_web(msg)


def _log_error(handler, error):
    """Logs an error response.

    """
    _log(handler, "error --> {} --> {}".format(handler, error), True)


def _log_success(handler):
    """Logs a successful response.

    """
    _log(handler, "success --> {}".format(handler))


def _write_null(handler, data):
    """Writes HTTP response null data.

    """
    pass


def _write_csv(handler, data):
    """Writes HTTP response CSV data.

    """
    handler.write(data)
    handler.set_header("Content-Type", "application/csv; charset=utf-8")


def _write_json(handler, data):
    """Writes HTTP response JSON data.

    """
    data['pyessvVersion'] = pyessv.__version__
    handler.write(to_dict(data, to_camel_case))
    handler.set_header("Content-Type", "application/json; charset=utf-8")


def _write_html(handler, data):
    """Writes HTTP response HTML data.

    """
    handler.write(data)
    handler.set_header("Content-Type", "text/html; charset=utf-8")


def _write_pdf(handler, data):
    """Writes HTTP response PDF data.

    """
    handler.write(data)
    handler.set_header("Content-Type", "application/pdf; charset=utf-8")


def _write_xml(handler, data):
    """Writes HTTP response XML data.

    """
    handler.write(data)
    handler.set_header("Content-Type", "application/xml; charset=utf-8")


# Map of response writers to encodings.
_WRITERS = {
    None: _write_null,
    'csv': _write_csv,
    'json': _write_json,
    'html': _write_html,
    'pdf': _write_pdf,
    'xml': _write_xml
}


def _write(handler, data, encoding='json'):
    """Writes HTTP response data.

    """
    # Log begin.
    _log(handler, "response writing begins --> {}".format(handler))

    # Write.
    _WRITERS[encoding](handler, data)

    # Log end.
    _log(handler, "response writing ends --> {}".format(handler))


def write_error(handler, error):
    """Writes processing error to response stream.

    """
    # Reset handler output.
    handler.clear()

    # Set error info to be returned to client.
    _write(handler, {
        'error_code': ERROR_CODES.get(type(error), 999),
        'error_field': getattr(error, 'field', '--'),
        'error_message': error.message.strip(),
        'error_type': type(error).__name__
    })

    # Set response HTTP status code.
    try:
        handler.set_status(error.response_code)
    except AttributeError:
        handler.set_status(HTTP_RESPONSE_SERVER_ERROR)


def _write_success(handler):
    """Writes processing success to response stream.

    """
    # Set response encoding.
    try:
        encoding = handler.output_encoding
    except AttributeError:
        encoding = 'json'

    # Set response data.
    try:
        data = handler.output
    except AttributeError:
        data = unicode()
        encoding = None

    # Write respponse.
    _write(handler, data, encoding)

    # Clean up.
    try:
        del handler.output
    except AttributeError:
        pass


def _get_tasks(pre_tasks, tasks, post_tasks):
    """Returns formatted & extended taskset.

    """
    try:
        iter(tasks)
    except TypeError:
        tasks = [tasks]

    return pre_tasks + tasks + post_tasks


def _invoke_task(handler, task, err=None):
    """Invokes a task.

    """
    try:
        if err:
            task(handler, err)
        else:
            task(handler)
    except TypeError as te:
        if err:
            task(err)
        else:
            task()



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._accessors.esdoc_cmip6.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Encapsulates access to ESDOC CMIP6 vocabs.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import pyessv



def _get_scope():
    """Returns target scope.

    """
    return pyessv.ESDOC.cmip6


def get_model_topics(source_id=None):
    """Returns topics to be documented by a source identifier.

    """
    def _requires_documentation(topic):
        if topic.canonical_name == 'toplevel':
            return True
        return source_id.data['model_component'][topic.raw_name]['description'] != 'none'

    scope = _get_scope()
    if source_id is None:
        return scope.model_topic
    else:
        return [i for i in scope.model_topic if _requires_documentation(i)]
